#!/usr/bin/env python
from sqlalchemy import Column, Float, Index, Integer, Text

from . import HummingbotBase


class RangePositionCollectedFees(HummingbotBase):
    """
    Table schema used when LP feesmare claimed.
    """
    __tablename__ = "RangePositionCollectedFees"
    __table_args__ = (Index("rpf_id_index",
                            "token_id", "config_file_path"),
                      )
    id = Column(Integer, primary_key=True, nullable=False)
    config_file_path = Column(Text, nullable=False)
    strategy = Column(Text, nullable=False)
    token_id = Column(Integer, nullable=False)
    token_0 = Column(Text, nullable=False)
    token_1 = Column(Text, nullable=False)
    claimed_fee_0 = Column(Float, nullable=False)
    claimed_fee_1 = Column(Float, nullable=False)

    def __repr__(self) -> str:
        return f"RangePositionCollectedFees(id={self.id}, config_file_path='{self.config_file_path}', strategy='{self.strategy}', " \
               f"token_id={self.token_id}, token_0='{self.token_0}', token_1='{self.token_1}', " \
               f"claimed_fee_0={self.claimed_fee_0}, claimed_fee_1={self.claimed_fee_1})"



from abc import ABC, abstractmethod
from decimal import Decimal
from typing import Dict, Tuple, Union

from pydantic import ConfigDict, field_validator, model_validator
from pydantic.v1 import Field, validator

import hummingbot.client.settings as settings
from hummingbot.client.config.config_data_types import BaseClientModel, ClientConfigEnum, ClientFieldData
from hummingbot.client.config.config_validators import validate_bool, validate_connector
from hummingbot.client.config.strategy_config_data_types import BaseTradingStrategyMakerTakerConfigMap
from hummingbot.client.settings import AllConnectorSettings
from hummingbot.core.data_type.trade_fee import TokenAmount
from hummingbot.core.rate_oracle.rate_oracle import RateOracle
from hummingbot.strategy.maker_taker_market_pair import MakerTakerMarketPair


class ConversionRateModel(BaseClientModel, ABC):
    @abstractmethod
    def get_conversion_rates(
        self, market_pair: MakerTakerMarketPair
    ) -> Tuple[str, str, Decimal, str, str, Decimal]:
        pass


class OracleConversionRateMode(ConversionRateModel):
    model_config = ConfigDict(title="rate_oracle_conversion_rate")

    def get_conversion_rates(
        self, market_pair: MakerTakerMarketPair
    ) -> Tuple[str, str, Decimal, str, str, Decimal]:
        """
        Find conversion rates from taker market to maker market
        :param market_pair: maker and taker trading pairs for which to do conversion
        :return: A tuple of quote pair symbol, quote conversion rate source, quote conversion rate,
        base pair symbol, base conversion rate source, base conversion rate
        """
        from .cross_exchange_market_making import CrossExchangeMarketMakingStrategy
        quote_pair = f"{market_pair.taker.quote_asset}-{market_pair.maker.quote_asset}"
        if market_pair.taker.quote_asset != market_pair.maker.quote_asset:
            quote_rate_source = RateOracle.get_instance().source.name
            quote_rate = RateOracle.get_instance().get_pair_rate(quote_pair)
        else:
            quote_rate_source = "fixed"
            quote_rate = Decimal("1")

        base_pair = f"{market_pair.taker.base_asset}-{market_pair.maker.base_asset}"
        if market_pair.taker.base_asset != market_pair.maker.base_asset:
            base_rate_source = RateOracle.get_instance().source.name
            base_rate = RateOracle.get_instance().get_pair_rate(base_pair)
        else:
            base_rate_source = "fixed"
            base_rate = Decimal("1")

        gas_pair = None
        if CrossExchangeMarketMakingStrategy.is_gateway_market(market_pair.taker):
            if hasattr(market_pair.taker.market, "network_transaction_fee"):
                transaction_fee: TokenAmount = market_pair.taker.market.network_transaction_fee
                if transaction_fee is not None:
                    gas_pair = f"{transaction_fee.token}-{market_pair.maker.quote_asset}"

        if gas_pair is not None and transaction_fee.token != market_pair.maker.quote_asset:
            gas_rate_source = RateOracle.get_instance().source.name
            gas_rate = RateOracle.get_instance().get_pair_rate(gas_pair)
        else:
            gas_rate_source = "fixed"
            gas_rate = Decimal("1")

        return quote_pair, quote_rate_source, quote_rate, base_pair, base_rate_source, base_rate, gas_pair, gas_rate_source, gas_rate


class TakerToMakerConversionRateMode(ConversionRateModel):
    taker_to_maker_base_conversion_rate: Decimal = Field(
        default=Decimal("1.0"),
        description="A fixed conversion rate between the maker and taker trading pairs based on the maker base asset.",
        gt=0.0,
        client_data=ClientFieldData(
            prompt=lambda mi: (
                "Enter conversion rate for taker base asset value to maker base asset value, e.g. "
                "if maker base asset is USD and the taker is DAI, 1 DAI is valued at 1.25 USD, "
                "the conversion rate is 1.25"
            ),
            prompt_on_new=True,
        ),
    )
    taker_to_maker_quote_conversion_rate: Decimal = Field(
        default=Decimal("1.0"),
        description="A fixed conversion rate between the maker and taker trading pairs based on the maker quote asset.",
        gt=0.0,
        client_data=ClientFieldData(
            prompt=lambda mi: (
                "Enter conversion rate for taker quote asset value to maker quote asset value, e.g. "
                "if maker quote asset is USD and the taker is DAI, 1 DAI is valued at 1.25 USD, "
                "the conversion rate is 1.25"
            ),
            prompt_on_new=True,
        ),
    )
    gas_to_maker_base_conversion_rate: Decimal = Field(
        default=Decimal("1.0"),
        description="A fixed conversion rate between the maker quote asset and taker gas asset.",
        gt=0.0,
        client_data=ClientFieldData(
            prompt=lambda mi: (
                "Enter conversion rate for gas token value of taker gateway exchange to maker base asset value, e.g. "
                "if maker base asset is USD and the gas token is DAI, 1 DAI is valued at 1.25 USD, "
                "the conversion rate is 1.25"
            ),
            prompt_on_new=True,
        ),
    )
    model_config = ConfigDict(title="fixed_conversion_rate")

    def get_conversion_rates(
        self, market_pair: MakerTakerMarketPair
    ) -> Tuple[str, str, Decimal, str, str, Decimal]:
        """
        Find conversion rates from taker market to maker market
        :param market_pair: maker and taker trading pairs for which to do conversion
        :return: A tuple of quote pair symbol, quote conversion rate source, quote conversion rate,
        base pair symbol, base conversion rate source, base conversion rate
        """
        from .cross_exchange_market_making import CrossExchangeMarketMakingStrategy
        quote_pair = f"{market_pair.taker.quote_asset}-{market_pair.maker.quote_asset}"
        quote_rate_source = "fixed"
        quote_rate = self.taker_to_maker_quote_conversion_rate

        base_pair = f"{market_pair.taker.base_asset}-{market_pair.maker.base_asset}"
        base_rate_source = "fixed"
        base_rate = self.taker_to_maker_base_conversion_rate

        gas_pair = None
        if CrossExchangeMarketMakingStrategy.is_gateway_market(market_pair.taker):
            if hasattr(market_pair.taker.market, "network_transaction_fee"):
                transaction_fee: TokenAmount = market_pair.taker.market.network_transaction_fee
                if transaction_fee is not None:
                    gas_pair = f"{transaction_fee.token}-{market_pair.maker.quote_asset}"

        gas_rate_source = "fixed"
        gas_rate = self.taker_to_maker_base_conversion_rate

        return quote_pair, quote_rate_source, quote_rate, base_pair, base_rate_source, base_rate, gas_pair, gas_rate_source, gas_rate


CONVERSION_RATE_MODELS = {
    OracleConversionRateMode.model_config["title"]: OracleConversionRateMode,
    TakerToMakerConversionRateMode.model_config["title"]: TakerToMakerConversionRateMode,
}


class OrderRefreshMode(BaseClientModel, ABC):
    @abstractmethod
    def get_cancel_order_threshold(self) -> Decimal:
        pass

    @abstractmethod
    def get_expiration_seconds(self) -> Decimal:
        pass


class PassiveOrderRefreshMode(OrderRefreshMode):
    cancel_order_threshold: Decimal = Field(
        default=Decimal("5.0"),
        description="Profitability threshold to cancel a trade.",
        gt=-100.0,
        lt=100.0,
        client_data=ClientFieldData(
            prompt=lambda mi: "What is the threshold of profitability to cancel a trade? (Enter 1 to indicate 1%)",
            prompt_on_new=True,
        ),
    )

    limit_order_min_expiration: Decimal = Field(
        default=130.0,
        description="Limit order expiration time limit.",
        gt=0.0,
        client_data=ClientFieldData(
            prompt=lambda mi: "How often do you want limit orders to expire (in seconds)?",
            prompt_on_new=True,
        ),
    )
    model_config = ConfigDict(title="passive_order_refresh")

    def get_cancel_order_threshold(self) -> Decimal:
        return self.cancel_order_threshold / Decimal("100")

    def get_expiration_seconds(self) -> Decimal:
        return self.limit_order_min_expiration


class ActiveOrderRefreshMode(OrderRefreshMode):
    model_config = ConfigDict(title="active_order_refresh")

    def get_cancel_order_threshold(self) -> Decimal:
        return Decimal('nan')

    def get_expiration_seconds(self) -> Decimal:
        return Decimal('nan')


ORDER_REFRESH_MODELS = {
    PassiveOrderRefreshMode.model_config["title"]: PassiveOrderRefreshMode,
    ActiveOrderRefreshMode.model_config["title"]: ActiveOrderRefreshMode,
}


class CrossExchangeMarketMakingConfigMap(BaseTradingStrategyMakerTakerConfigMap):
    strategy: str = Field(default="cross_exchange_market_making", client_data=None)

    min_profitability: Decimal = Field(
        default=...,
        description="The minimum estimated profitability required to open a position.",
        ge=-100.0,
        le=100.0,
        client_data=ClientFieldData(
            prompt=lambda mi: "What is the minimum profitability for you to make a trade? (Enter 1 to indicate 1%)",
            prompt_on_new=True,
        ),
    )
    order_amount: Decimal = Field(
        default=...,
        description="The strategy order amount.",
        ge=0.0,
        client_data=ClientFieldData(
            prompt=lambda mi: CrossExchangeMarketMakingConfigMap.order_amount_prompt(mi),
            prompt_on_new=True,
        )
    )
    adjust_order_enabled: bool = Field(
        default=True,
        description="Adjust order price to be one tick above the top bid or below the top ask.",
        client_data=ClientFieldData(
            prompt=lambda mi: "Do you want to enable adjust order? (Yes/No)"
        ),
    )
    order_refresh_mode: Union[ActiveOrderRefreshMode, PassiveOrderRefreshMode] = Field(
        default=ActiveOrderRefreshMode.construct(),
        description="Refresh orders by cancellation or by letting them expire.",
        client_data=ClientFieldData(
            prompt=lambda mi: f"Select the order refresh mode ({'/'.join(list(ORDER_REFRESH_MODELS.keys()))})",
            prompt_on_new=True,
        ),
    )
    top_depth_tolerance: Decimal = Field(
        default=Decimal("0.0"),
        description="Volume requirement for determining a possible top bid or ask price from the order book.",
        ge=0.0,
        client_data=ClientFieldData(
            prompt=lambda mi: CrossExchangeMarketMakingConfigMap.top_depth_tolerance_prompt(mi),
        ),
    )
    anti_hysteresis_duration: float = Field(
        default=60.0,
        description="Minimum time limit between two subsequent order adjustments.",
        gt=0.0,
        client_data=ClientFieldData(
            prompt=lambda mi: "What is the minimum time interval you want limit orders to be adjusted? (in seconds)",
        ),
    )
    order_size_taker_volume_factor: Decimal = Field(
        default=Decimal("25.0"),
        description="Taker order size as a percentage of volume.",
        ge=0.0,
        le=100.0,
        client_data=ClientFieldData(
            prompt=lambda mi: (
                "What percentage of hedge-able volume would you like to be traded on the taker market? "
                "(Enter 1 to indicate 1%)"
            ),
        ),
    )
    order_size_taker_balance_factor: Decimal = Field(
        default=Decimal("99.5"),
        description="Taker order size as a percentage of the available balance.",
        ge=0.0,
        le=100.0,
        client_data=ClientFieldData(
            prompt=lambda mi: (
                "What percentage of asset balance would you like to use for hedging trades on the taker market? "
                "(Enter 1 to indicate 1%)"
            ),
        ),
    )
    order_size_portfolio_ratio_limit: Decimal = Field(
        default=Decimal("16.67"),
        description="Order size as a maker and taker account balance ratio.",
        ge=0.0,
        le=100.0,
        client_data=ClientFieldData(
            prompt=lambda mi: (
                "What ratio of your total portfolio value would you like to trade on the maker and taker markets? "
                "Enter 50 for 50%"
            ),
        ),
    )
    conversion_rate_mode: Union[OracleConversionRateMode, TakerToMakerConversionRateMode] = Field(
        default=OracleConversionRateMode.construct(),
        description="Convert between different trading pairs using fixed conversion rates or using the rate oracle.",
        client_data=ClientFieldData(
            prompt=lambda mi: f"Select the conversion rate mode ({'/'.join(list(CONVERSION_RATE_MODELS.keys()))})",
            prompt_on_new=True,
        ),
    )
    slippage_buffer: Decimal = Field(
        default=Decimal("5.0"),
        description="Allowed slippage to fill ensure taker orders are filled.",
        ge=0.0,
        le=100.0,
        client_data=ClientFieldData(
            prompt=lambda mi: (
                "How much buffer do you want to add to the price to account for slippage for taker orders "
                "Enter 1 to indicate 1%"
            ),
            prompt_on_new=True,
        ),
    )

    debug_price_shim: bool = Field(
        default=False,
        description="Usd the debug price shim to mock gateway price.",
        client_data=ClientFieldData(
            prompt=lambda mi: (
                "Do you want to enable the debug price shim for integration tests? If you don't know what this does "
                "you should keep it disabled."
            ),
        ),
    )
    gateway_transaction_cancel_interval: int = Field(
        default= 600,
        description="Gateway transaction cancellation timeout.",
        ge=1,
        client_data=ClientFieldData(
            prompt=lambda mi: (
                "After what time should blockchain transactions be cancelled if they are not included in a block? "
                "(this only affects decentralized exchanges) (Enter time in seconds)"
            ),
            prompt_on_new=True,
        ),
    )
    taker_market: ClientConfigEnum(
        value="TakerMarkets",  # noqa: F821
        names={e: e for e in
               sorted(AllConnectorSettings.get_exchange_names().union(
                   AllConnectorSettings.get_gateway_amm_connector_names()
               ))},
        type=str,
    ) = Field(
        default=...,
        description="The name of the taker exchange connector.",
        client_data=ClientFieldData(
            prompt=lambda mi: "Enter your taker connector (Exchange/AMM/CLOB)",
            prompt_on_new=True,
        ),
    )

    # === prompts ===

    @classmethod
    def top_depth_tolerance_prompt(cls, model_instance: 'CrossExchangeMarketMakingConfigMap') -> str:
        maker_market = model_instance.maker_market_trading_pair
        base_asset, quote_asset = maker_market.split("-")
        return f"What is your top depth tolerance? (in {base_asset})"

    @classmethod
    def order_amount_prompt(cls, model_instance: 'CrossExchangeMarketMakingConfigMap') -> str:
        trading_pair = model_instance.maker_market_trading_pair
        base_asset, quote_asset = trading_pair.split("-")
        return f"What is the amount of {base_asset} per order?"

    # === specific validations ===
    @field_validator("order_refresh_mode", mode="before")
    @classmethod
    def validate_order_refresh_mode(cls, v: Union[str, ActiveOrderRefreshMode, PassiveOrderRefreshMode]):
        if isinstance(v, (ActiveOrderRefreshMode, PassiveOrderRefreshMode, Dict)):
            sub_model = v
        elif v not in ORDER_REFRESH_MODELS:
            raise ValueError(
                f"Invalid order refresh mode, please choose value from {list(ORDER_REFRESH_MODELS.keys())}."
            )
        else:
            sub_model = ORDER_REFRESH_MODELS[v].construct()
        return sub_model

    @field_validator("conversion_rate_mode", mode="before")
    @classmethod
    def validate_conversion_rate_mode(cls, v: Union[str, OracleConversionRateMode, TakerToMakerConversionRateMode]):
        if isinstance(v, (OracleConversionRateMode, TakerToMakerConversionRateMode, Dict)):
            sub_model = v
        elif v not in CONVERSION_RATE_MODELS:
            raise ValueError(
                f"Invalid conversion rate mode, please choose value from {list(CONVERSION_RATE_MODELS.keys())}."
            )
        else:
            sub_model = CONVERSION_RATE_MODELS[v].construct()
        return sub_model

    # === generic validations ===

    @field_validator(
        "adjust_order_enabled",
        mode="before")
    @classmethod
    def validate_bool(cls, v: str):
        """Used for client-friendly error output."""
        if isinstance(v, str):
            ret = validate_bool(v)
            if ret is not None:
                raise ValueError(ret)
        return v

    # TODO[pydantic]: We couldn't refactor the `validator`, please replace it by `field_validator` manually.
    # Check https://docs.pydantic.dev/dev-v2/migration/#changes-to-validators for more information.
    @validator(
        "min_profitability",
        "order_amount",
        "top_depth_tolerance",
        "anti_hysteresis_duration",
        "order_size_taker_volume_factor",
        "order_size_taker_balance_factor",
        "order_size_portfolio_ratio_limit",
        "slippage_buffer",
        pre=True,
    )
    def validate_decimal(cls, v: str, field: Field):
        """Used for client-friendly error output."""
        return super().validate_decimal(v, field)

    # === post-validations ===

    @model_validator(mode="after")
    def post_validations(self):
        # Add the maker and taker markets to the required exchanges
        settings.required_exchanges.add(self["maker_market"])
        settings.required_exchanges.add(self["taker_market"])

        # if the conversion rate mode is oracle, we need to set the oracle pairs
        use_oracle = self.use_oracle_conversion_rate
        first_base, first_quote = self.maker_market_trading_pair.split("-")
        second_base, second_quote = self.taker_market_trading_pair.split("-")
        if use_oracle and (first_base != second_base or first_quote != second_quote):
            settings.required_rate_oracle = True
            settings.rate_oracle_pairs = []
            if first_base != second_base:
                settings.rate_oracle_pairs.append(f"{second_base}-{first_base}")
            if first_quote != second_quote:
                settings.rate_oracle_pairs.append(f"{second_quote}-{first_quote}")
        else:
            settings.required_rate_oracle = False
            settings.rate_oracle_pairs = []
        return self

    # TODO[pydantic]: We couldn't refactor the `validator`, please replace it by `field_validator` manually.
    # Check https://docs.pydantic.dev/dev-v2/migration/#changes-to-validators for more information.
    @field_validator("maker_market", "taker_market", mode="before")
    def validate_exchange(cls, v: str, field: Field):
        """Used for client-friendly error output."""
        if field.name == "maker_market":
            super().validate_exchange(v=v, field=field)
        if field.name == "taker_market":
            ret = validate_connector(v)
            if ret is not None:
                raise ValueError(ret)
            cls.__fields__["taker_market"].type_ = ClientConfigEnum(  # rebuild the exchanges enum
                value="TakerMarkets",  # noqa: F821
                names={e: e for e in sorted(
                    AllConnectorSettings.get_exchange_names().union(
                        AllConnectorSettings.get_gateway_amm_connector_names()
                    ))},
                type=str,
            )
            cls._clear_schema_cache()
        return v



import asyncio
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple

from bidict import bidict

from hummingbot.connector.constants import s_decimal_NaN
from hummingbot.connector.exchange.kucoin import (
    kucoin_constants as CONSTANTS,
    kucoin_utils as utils,
    kucoin_web_utils as web_utils,
)
from hummingbot.connector.exchange.kucoin.kucoin_api_order_book_data_source import KucoinAPIOrderBookDataSource
from hummingbot.connector.exchange.kucoin.kucoin_api_user_stream_data_source import KucoinAPIUserStreamDataSource
from hummingbot.connector.exchange.kucoin.kucoin_auth import KucoinAuth
from hummingbot.connector.exchange_py_base import ExchangePyBase
from hummingbot.connector.trading_rule import TradingRule
from hummingbot.connector.utils import combine_to_hb_trading_pair
from hummingbot.core.data_type.common import OrderType, TradeType
from hummingbot.core.data_type.in_flight_order import InFlightOrder, OrderState, OrderUpdate, TradeUpdate
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.data_type.trade_fee import AddedToCostTradeFee, TokenAmount, TradeFeeBase
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.utils.estimate_fee import build_trade_fee
from hummingbot.core.web_assistant.connections.data_types import RESTMethod
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory

if TYPE_CHECKING:
    from hummingbot.client.config.config_helpers import ClientConfigAdapter


class KucoinExchange(ExchangePyBase):
    web_utils = web_utils

    def __init__(self,
                 client_config_map: "ClientConfigAdapter",
                 kucoin_api_key: str,
                 kucoin_passphrase: str,
                 kucoin_secret_key: str,
                 trading_pairs: Optional[List[str]] = None,
                 trading_required: bool = True,
                 domain: str = CONSTANTS.DEFAULT_DOMAIN):
        self.kucoin_api_key = kucoin_api_key
        self.kucoin_passphrase = kucoin_passphrase
        self.kucoin_secret_key = kucoin_secret_key
        self._domain = domain
        self._trading_required = trading_required
        self._trading_pairs = trading_pairs
        self._last_order_fill_ts_s: float = 0
        super().__init__(client_config_map=client_config_map)

    @property
    def authenticator(self):
        return KucoinAuth(
            api_key=self.kucoin_api_key,
            passphrase=self.kucoin_passphrase,
            secret_key=self.kucoin_secret_key,
            time_provider=self._time_synchronizer)

    @property
    def name(self) -> str:
        return "kucoin"

    @property
    def rate_limits_rules(self):
        return CONSTANTS.RATE_LIMITS

    @property
    def domain(self):
        return self._domain

    @property
    def client_order_id_max_length(self):
        return CONSTANTS.MAX_ORDER_ID_LEN

    @property
    def client_order_id_prefix(self):
        return ""

    @property
    def trading_rules_request_path(self):
        return CONSTANTS.SYMBOLS_PATH_URL

    @property
    def trading_pairs_request_path(self):
        return CONSTANTS.SYMBOLS_PATH_URL

    @property
    def check_network_request_path(self):
        return CONSTANTS.SERVER_TIME_PATH_URL

    @property
    def orders_path_url(self):
        return CONSTANTS.ORDERS_PATH_URL_HFT if self._domain == "hft" else CONSTANTS.ORDERS_PATH_URL

    @property
    def fills_path_url(self):
        return CONSTANTS.FILLS_PATH_URL_HFT if self.domain == "hft" else CONSTANTS.FILLS_PATH_URL

    @property
    def trading_pairs(self):
        return self._trading_pairs

    @property
    def is_cancel_request_in_exchange_synchronous(self) -> bool:
        return True

    @property
    def is_trading_required(self) -> bool:
        return self._trading_required

    def supported_order_types(self):
        return [OrderType.MARKET, OrderType.LIMIT, OrderType.LIMIT_MAKER]

    async def get_all_pairs_prices(self) -> List[Dict[str, str]]:
        pairs_prices = await self._api_get(path_url=CONSTANTS.ALL_TICKERS_PATH_URL)
        return pairs_prices

    def _is_request_exception_related_to_time_synchronizer(self, request_exception: Exception):
        error_description = str(request_exception)
        return CONSTANTS.RET_CODE_AUTH_TIMESTAMP_ERROR in error_description and CONSTANTS.RET_MSG_AUTH_TIMESTAMP_ERROR in error_description

    def _is_order_not_found_during_status_update_error(self, status_update_exception: Exception) -> bool:
        return (str(CONSTANTS.RET_CODE_RESOURCE_NOT_FOUND) in str(status_update_exception) and
                str(CONSTANTS.RET_MSG_RESOURCE_NOT_FOUND) in str(status_update_exception))

    def _is_order_not_found_during_cancelation_error(self, cancelation_exception: Exception) -> bool:
        return (str(CONSTANTS.RET_CODE_ORDER_NOT_EXIST_OR_NOT_ALLOW_TO_CANCEL) in str(cancelation_exception)
                and str(CONSTANTS.RET_MSG_ORDER_NOT_EXIST_OR_NOT_ALLOW_TO_CANCEL) in str(cancelation_exception))

    def _create_web_assistants_factory(self) -> WebAssistantsFactory:
        return web_utils.build_api_factory(
            throttler=self._throttler,
            time_synchronizer=self._time_synchronizer,
            domain=self.domain,
            auth=self._auth)

    def _create_order_book_data_source(self) -> OrderBookTrackerDataSource:
        return KucoinAPIOrderBookDataSource(
            trading_pairs=self._trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory,
            domain=self.domain,
        )

    def _create_user_stream_data_source(self) -> UserStreamTrackerDataSource:
        return KucoinAPIUserStreamDataSource(
            auth=self._auth,
            trading_pairs=self._trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory,
            domain=self.domain,
        )

    def _get_fee(self,
                 base_currency: str,
                 quote_currency: str,
                 order_type: OrderType,
                 order_side: TradeType,
                 amount: Decimal,
                 price: Decimal = s_decimal_NaN,
                 is_maker: Optional[bool] = None) -> AddedToCostTradeFee:

        is_maker = is_maker or (order_type is OrderType.LIMIT_MAKER)
        trading_pair = combine_to_hb_trading_pair(base=base_currency, quote=quote_currency)
        if trading_pair in self._trading_fees:
            fees_data = self._trading_fees[trading_pair]
            fee_value = Decimal(fees_data["makerFeeRate"]) if is_maker else Decimal(fees_data["takerFeeRate"])
            fee = AddedToCostTradeFee(percent=fee_value)
        else:
            fee = build_trade_fee(
                self.name,
                is_maker,
                base_currency=base_currency,
                quote_currency=quote_currency,
                order_type=order_type,
                order_side=order_side,
                amount=amount,
                price=price,
            )
        return fee

    def _initialize_trading_pair_symbols_from_exchange_info(self, exchange_info: Dict[str, Any]):
        mapping = bidict()
        for symbol_data in filter(utils.is_pair_information_valid, exchange_info.get("data", [])):
            mapping[symbol_data["symbol"]] = combine_to_hb_trading_pair(base=symbol_data["baseCurrency"],
                                                                        quote=symbol_data["quoteCurrency"])
        self._set_trading_pair_symbol_map(mapping)

    async def _place_order(self,
                           order_id: str,
                           trading_pair: str,
                           amount: Decimal,
                           trade_type: TradeType,
                           order_type: OrderType,
                           price: Decimal,
                           **kwargs) -> Tuple[str, float]:
        side = trade_type.name.lower()
        order_type_str = "market" if order_type == OrderType.MARKET else "limit"
        data = {
            "size": str(amount),
            "clientOid": order_id,
            "side": side,
            "symbol": await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair),
            "type": order_type_str,
        }
        if order_type is OrderType.LIMIT:
            data["price"] = str(price)
        elif order_type is OrderType.LIMIT_MAKER:
            data["price"] = str(price)
            data["postOnly"] = True
        exchange_order_id = await self._api_post(
            path_url=self.orders_path_url,
            data=data,
            is_auth_required=True,
            limit_id=CONSTANTS.POST_ORDER_LIMIT_ID,
        )
        if exchange_order_id.get("data") is None:
            raise IOError(f"Error placing order on Kucoin: {exchange_order_id}")
        return str(exchange_order_id["data"]["orderId"]), self.current_timestamp

    async def _place_cancel(self, order_id: str, tracked_order: InFlightOrder):
        """
        This implementation specific function is called by _cancel, and returns True if successful
        """
        exchange_order_id = await tracked_order.get_exchange_order_id()
        params = {"symbol": tracked_order.trading_pair} if self.domain == "hft" else None
        cancel_result = await self._api_delete(
            f"{self.orders_path_url}/{exchange_order_id}",
            params=params,
            is_auth_required=True,
            limit_id=CONSTANTS.DELETE_ORDER_LIMIT_ID
        )
        response_param = "orderId" if self.domain == "hft" else "cancelledOrderIds"
        if cancel_result.get("data") is not None:
            return tracked_order.exchange_order_id in cancel_result["data"].get(response_param, [])
        else:
            raise IOError(f"Error cancelling order on Kucoin: {cancel_result}")

    async def _user_stream_event_listener(self):
        """
        This functions runs in background continuously processing the events received from the exchange by the user
        stream data source. It keeps reading events from the queue until the task is interrupted.
        The events received are balance updates, order updates and trade events.
        """
        async for event_message in self._iter_user_event_queue():
            try:
                event_type = event_message.get("type")
                event_subject = event_message.get("subject")
                execution_data = event_message.get("data")

                # Refer to https://docs.kucoin.com/#private-order-change-events
                if event_type == "message" and event_subject == CONSTANTS.ORDER_CHANGE_EVENT_TYPE:
                    order_event_type = execution_data["type"]
                    client_order_id: Optional[str] = execution_data.get("clientOid")

                    fillable_order = self._order_tracker.all_fillable_orders.get(client_order_id)
                    updatable_order = self._order_tracker.all_updatable_orders.get(client_order_id)

                    event_timestamp = execution_data["ts"] * 1e-9

                    if fillable_order is not None and order_event_type == "match":
                        execute_amount_diff = Decimal(execution_data["matchSize"])
                        execute_price = Decimal(execution_data["matchPrice"])

                        fee = self.get_fee(
                            fillable_order.base_asset,
                            fillable_order.quote_asset,
                            fillable_order.order_type,
                            fillable_order.trade_type,
                            execute_price,
                            execute_amount_diff,
                        )

                        trade_update = TradeUpdate(
                            trade_id=execution_data["tradeId"],
                            client_order_id=client_order_id,
                            exchange_order_id=execution_data["orderId"],
                            trading_pair=updatable_order.trading_pair,
                            fee=fee,
                            fill_base_amount=execute_amount_diff,
                            fill_quote_amount=execute_amount_diff * execute_price,
                            fill_price=execute_price,
                            fill_timestamp=event_timestamp,
                        )
                        self._order_tracker.process_trade_update(trade_update)

                    if updatable_order is not None:
                        updated_status = updatable_order.current_state
                        if order_event_type == "open":
                            updated_status = OrderState.OPEN
                        elif order_event_type == "match":
                            updated_status = OrderState.PARTIALLY_FILLED
                        elif order_event_type == "filled":
                            updated_status = OrderState.FILLED
                        elif order_event_type == "canceled":
                            updated_status = OrderState.CANCELED

                        order_update = OrderUpdate(
                            trading_pair=updatable_order.trading_pair,
                            update_timestamp=event_timestamp,
                            new_state=updated_status,
                            client_order_id=client_order_id,
                            exchange_order_id=execution_data["orderId"],
                        )
                        self._order_tracker.process_order_update(order_update=order_update)

                elif event_type == "message" and event_subject == CONSTANTS.BALANCE_EVENT_TYPE:
                    currency = execution_data["currency"]
                    available_balance = Decimal(execution_data["available"])
                    total_balance = Decimal(execution_data["total"])
                    self._account_balances.update({currency: total_balance})
                    self._account_available_balances.update({currency: available_balance})

            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().exception("Unexpected error in user stream listener loop.")
                await self._sleep(5.0)

    async def _update_balances(self):
        local_asset_names = set(self._account_balances.keys())
        remote_asset_names = set()
        account_type = "trade_hf" if self.domain == "hft" else "trade"

        response = await self._api_get(
            path_url=CONSTANTS.ACCOUNTS_PATH_URL,
            params={"type": account_type},
            is_auth_required=True)

        if response:
            for balance_entry in response["data"]:
                asset_name = balance_entry["currency"]
                self._account_available_balances[asset_name] = Decimal(balance_entry["available"])
                self._account_balances[asset_name] = Decimal(balance_entry["balance"])
                remote_asset_names.add(asset_name)

            asset_names_to_remove = local_asset_names.difference(remote_asset_names)
            for asset_name in asset_names_to_remove:
                del self._account_available_balances[asset_name]
                del self._account_balances[asset_name]

    async def _format_trading_rules(self, raw_trading_pair_info: Dict[str, Any]) -> List[TradingRule]:
        trading_rules = []

        for info in raw_trading_pair_info["data"]:
            if utils.is_pair_information_valid(info):
                try:
                    trading_pair = await self.trading_pair_associated_to_exchange_symbol(symbol=info.get("symbol"))
                    trading_rules.append(
                        TradingRule(trading_pair=trading_pair,
                                    min_order_size=Decimal(info["baseMinSize"]),
                                    max_order_size=Decimal(info["baseMaxSize"]),
                                    min_price_increment=Decimal(info['priceIncrement']),
                                    min_base_amount_increment=Decimal(info['baseIncrement']),
                                    min_quote_amount_increment=Decimal(info['quoteIncrement']),
                                    min_notional_size=Decimal(info["quoteMinSize"]))
                    )
                except Exception:
                    self.logger().error(f"Error parsing the trading pair rule {info}. Skipping.", exc_info=True)
        return trading_rules

    async def _update_trading_fees(self):
        trading_symbols = [await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
                           for trading_pair in self._trading_pairs]
        fees_json = []
        for idx in range(0, len(trading_symbols), CONSTANTS.TRADING_FEES_SYMBOL_LIMIT):
            sub_trading_symbols = trading_symbols[idx:idx + CONSTANTS.TRADING_FEES_SYMBOL_LIMIT]
            params = {"symbols": ",".join(sub_trading_symbols)}
            resp = await self._api_get(
                path_url=CONSTANTS.FEE_PATH_URL,
                params=params,
                is_auth_required=True,
            )
            fees_json.extend(resp["data"])

        for fee_json in fees_json:
            trading_pair = await self.trading_pair_associated_to_exchange_symbol(symbol=fee_json["symbol"])
            self._trading_fees[trading_pair] = fee_json

    async def _update_orders_fills(self, orders: List[InFlightOrder]):
        # This method in the base ExchangePyBase, makes an API call for each order.
        # Given the rate limit of the API method and the breadth of info provided by the method
        # the mitigation proposal is to collect all orders in one shot, then parse them
        # Note that this is limited to 500 orders (pagination)
        # An alternative for Kucoin would be to use the limit/fills that returns 24hr updates, which should
        # be sufficient, the rate limit seems better suited
        all_trades_updates: List[TradeUpdate] = []
        if len(orders) > 0:
            try:
                all_trades_updates: List[TradeUpdate] = await self._all_trades_updates(orders)
            except asyncio.CancelledError:
                raise
            except Exception as request_error:
                self.logger().warning(
                    f"Failed to fetch trade updates. Error: {request_error}")

            for trade_update in all_trades_updates:
                self._order_tracker.process_trade_update(trade_update)

    async def _all_trades_updates(self, orders: List[InFlightOrder]) -> List[TradeUpdate]:
        trade_updates: List[TradeUpdate] = []
        if len(orders) > 0:
            exchange_to_client = {o.exchange_order_id: {"client_id": o.client_order_id, "trading_pair": o.trading_pair} for o in orders}

            # We request updates from either:
            #    - The earliest order creation_timestamp in the list (first couple requests)
            #    - The last time we got a fill
            self._last_order_fill_ts_s = int(max(self._last_order_fill_ts_s, min([o.creation_timestamp for o in orders])))

            # From Kucoin https://docs.kucoin.com/#list-fills:
            # "If you only specified the start time, the system will automatically
            #  calculate the end time (end time = start time + 7 * 24 hours)"
            all_fills_response = await self._api_get(
                path_url=self.fills_path_url,
                params={
                    "pageSize": 500,
                    "startAt": self._last_order_fill_ts_s * 1000,
                },
                is_auth_required=True)

            for trade in all_fills_response.get("items", []):
                if str(trade["orderId"]) in exchange_to_client:
                    fee = TradeFeeBase.new_spot_fee(
                        fee_schema=self.trade_fee_schema(),
                        trade_type=TradeType.BUY if trade["side"] == "buy" else "sell",
                        percent_token=trade["feeCurrency"],
                        flat_fees=[TokenAmount(amount=Decimal(trade["fee"]), token=trade["feeCurrency"])]
                    )

                    client_info = exchange_to_client[str(trade["orderId"])]
                    trade_update = TradeUpdate(
                        trade_id=str(trade["tradeId"]),
                        client_order_id=client_info["client_id"],
                        trading_pair=client_info["trading_pair"],
                        exchange_order_id=str(trade["orderId"]),
                        fee=fee,
                        fill_base_amount=Decimal(trade["size"]),
                        fill_quote_amount=Decimal(trade["funds"]),
                        fill_price=Decimal(trade["price"]),
                        fill_timestamp=trade["createdAt"] * 1e-3,
                    )
                    trade_updates.append(trade_update)
                    # Update the last fill timestamp with the latest one
                    self._last_order_fill_ts_s = max(self._last_order_fill_ts_s, trade["createdAt"] * 1e-3)

        return trade_updates

    async def _all_trade_updates_for_order(self, order: InFlightOrder) -> List[TradeUpdate]:
        raise Exception("Developer: This method should not be called, it is obsoleted for Kucoin")

        trade_updates = []

        if order.exchange_order_id is not None:
            exchange_order_id = order.exchange_order_id
            all_fills_response = await self._api_get(
                path_url=self.fills_path_url,
                params={
                    "orderId": exchange_order_id,
                    "pageSize": 500,
                },
                is_auth_required=True)

            for trade in all_fills_response.get("items", []):
                fee = TradeFeeBase.new_spot_fee(
                    fee_schema=self.trade_fee_schema(),
                    trade_type=order.trade_type,
                    percent_token=trade["feeCurrency"],
                    flat_fees=[TokenAmount(amount=Decimal(trade["fee"]), token=trade["feeCurrency"])]
                )
                trade_update = TradeUpdate(
                    trade_id=str(trade["tradeId"]),
                    client_order_id=order.client_order_id,
                    exchange_order_id=exchange_order_id,
                    trading_pair=order.trading_pair,
                    fee=fee,
                    fill_base_amount=Decimal(trade["size"]),
                    fill_quote_amount=Decimal(trade["funds"]),
                    fill_price=Decimal(trade["price"]),
                    fill_timestamp=trade["createdAt"] * 1e-3,
                )
                trade_updates.append(trade_update)

        return trade_updates

    async def _request_order_status(self, tracked_order: InFlightOrder) -> OrderUpdate:
        exchange_order_id = await tracked_order.get_exchange_order_id()
        params = {"symbol": tracked_order.trading_pair} if self.domain == "hft" else None
        updated_order_data = await self._api_get(
            path_url=f"{self.orders_path_url}/{exchange_order_id}",
            is_auth_required=True,
            params=params,
            limit_id=CONSTANTS.GET_ORDER_LIMIT_ID)

        ordered_canceled = updated_order_data["data"]["cancelExist"]
        is_active = updated_order_data["data"]["active"] if self.domain == "hft" else updated_order_data["data"]["isActive"]
        op_type = updated_order_data["data"]["opType"]

        new_state = tracked_order.current_state
        if ordered_canceled or op_type == "CANCEL":
            new_state = OrderState.CANCELED
        elif not is_active:
            new_state = OrderState.FILLED

        order_update = OrderUpdate(
            client_order_id=tracked_order.client_order_id,
            exchange_order_id=updated_order_data["data"]["id"],
            trading_pair=tracked_order.trading_pair,
            update_timestamp=self.current_timestamp,
            new_state=new_state,
        )

        return order_update

    async def _get_last_traded_price(self, trading_pair: str) -> float:
        params = {
            "symbol": await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        }

        resp_json = await self._api_request(
            path_url=CONSTANTS.TICKER_PRICE_CHANGE_PATH_URL,
            method=RESTMethod.GET,
            params=params
        )

        return float(resp_json["data"]["price"])



import asyncio
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Dict, Optional

from hummingbot.connector.utils import combine_to_hb_trading_pair
from hummingbot.core.network_iterator import NetworkStatus, safe_ensure_future
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, RESTRequest, RESTResponse
from hummingbot.core.web_assistant.rest_pre_processors import RESTPreProcessorBase
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.data_feed.coin_cap_data_feed import coin_cap_constants as CONSTANTS
from hummingbot.data_feed.data_feed_base import DataFeedBase
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.core.api_throttler.async_throttler import AsyncThrottler


class CoinCapAPIKeyAppender(RESTPreProcessorBase):
    def __init__(self, api_key: str):
        super().__init__()
        self._api_key = api_key

    async def pre_process(self, request: RESTRequest) -> RESTRequest:
        request.headers = request.headers or {}
        request.headers["Authorization"] = self._api_key
        return request


class CoinCapDataFeed(DataFeedBase):
    _logger: Optional[HummingbotLogger] = None
    _async_throttler: Optional["AsyncThrottler"] = None

    @classmethod
    def _get_async_throttler(cls) -> "AsyncThrottler":
        """This avoids circular imports."""
        from hummingbot.core.api_throttler.async_throttler import AsyncThrottler

        if cls._async_throttler is None:
            cls._async_throttler = AsyncThrottler(CONSTANTS.RATE_LIMITS)
        return cls._async_throttler

    def __init__(self, assets_map: Dict[str, str], api_key: str):
        super().__init__()
        self._assets_map = assets_map
        self._price_dict: Dict[str, Decimal] = {}
        self._api_factory: Optional[WebAssistantsFactory] = None
        self._api_key = api_key
        self._is_api_key_authorized = True
        self._prices_stream_task: Optional[asyncio.Task] = None

        self._ready_event.set()

    @property
    def name(self):
        return "coin_cap_api"

    @property
    def health_check_endpoint(self):
        return f"{CONSTANTS.BASE_REST_URL}{CONSTANTS.HEALTH_CHECK_ENDPOINT}"

    @property
    def universal_quote_token(self) -> str:
        return CONSTANTS.UNIVERSAL_QUOTE_TOKEN

    async def start_network(self):
        self._prices_stream_task = safe_ensure_future(self._stream_prices())

    async def stop_network(self):
        self._prices_stream_task and self._prices_stream_task.cancel()
        self._prices_stream_task = None

    async def check_network(self) -> NetworkStatus:
        try:
            await self._make_request(url=self.health_check_endpoint)
        except asyncio.CancelledError:
            raise
        except Exception:
            return NetworkStatus.NOT_CONNECTED
        return NetworkStatus.CONNECTED

    async def get_all_usd_quoted_prices(self) -> Dict[str, Decimal]:
        prices = (
            self._price_dict
            if self._prices_stream_task and len(self._price_dict) != 0
            else await self._get_all_usd_quoted_prices_by_rest_request()
        )
        return prices

    def _get_api_factory(self) -> WebAssistantsFactory:
        # Avoids circular logic (i.e. CoinCap needs a throttler, which needs a client config map, which needs
        # a data feed — CoinCap, in this case)
        if self._api_factory is None:
            self._api_factory = WebAssistantsFactory(
                throttler=self._get_async_throttler(),
                rest_pre_processors=[CoinCapAPIKeyAppender(api_key=self._api_key)],
            )
        return self._api_factory

    async def _get_all_usd_quoted_prices_by_rest_request(self) -> Dict[str, Decimal]:
        prices = {}
        url = f"{CONSTANTS.BASE_REST_URL}{CONSTANTS.ALL_ASSETS_ENDPOINT}"

        params = {
            "ids": ",".join(self._assets_map.values()),
        }

        data = await self._make_request(url=url, params=params)
        for asset_data in data["data"]:
            base = asset_data["symbol"]
            trading_pair = combine_to_hb_trading_pair(base=base, quote=CONSTANTS.UNIVERSAL_QUOTE_TOKEN)
            try:
                prices[trading_pair] = Decimal(asset_data["priceUsd"])
            except TypeError:
                continue

        return prices

    async def _make_request(self, url: str, params: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
        api_factory = self._get_api_factory()
        rest_assistant = await api_factory.get_rest_assistant()
        rate_limit_id = CONSTANTS.API_KEY_LIMIT_ID if self._is_api_key_authorized else CONSTANTS.NO_KEY_LIMIT_ID
        response = await rest_assistant.execute_request_and_get_response(
            url=url,
            throttler_limit_id=rate_limit_id,
            params=params,
            method=RESTMethod.GET,
        )
        self._check_is_api_key_authorized(response=response)
        data = await response.json()
        return data

    def _check_is_api_key_authorized(self, response: RESTResponse):
        self.logger().debug(f"CoinCap REST response headers: {response.headers}")
        self._is_api_key_authorized = int(response.headers["X-Ratelimit-Limit"]) == CONSTANTS.API_KEY_LIMIT
        if not self._is_api_key_authorized and self._api_key != "":
            self.logger().warning("CoinCap API key is not authorized. Please check your API key.")

    async def _stream_prices(self):
        while True:
            try:
                api_factory = self._get_api_factory()
                self._price_dict = await self._get_all_usd_quoted_prices_by_rest_request()
                ws = await api_factory.get_ws_assistant()
                symbols_map = {asset_id: symbol for symbol, asset_id in self._assets_map.items()}
                ws_url = f"{CONSTANTS.BASE_WS_URL}{','.join(self._assets_map.values())}"
                async with api_factory.throttler.execute_task(limit_id=CONSTANTS.WS_CONNECTIONS_LIMIT_ID):
                    await ws.connect(ws_url=ws_url)
                async for msg in ws.iter_messages():
                    for asset_id, price_str in msg.data.items():
                        base = symbols_map[asset_id]
                        trading_pair = combine_to_hb_trading_pair(base=base, quote=CONSTANTS.UNIVERSAL_QUOTE_TOKEN)
                        self._price_dict[trading_pair] = Decimal(price_str)
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().network(
                    log_msg="Unexpected error while streaming prices. Restarting the stream.",
                    exc_info=True,
                )
                await self._sleep(delay=1)

    @staticmethod
    async def _sleep(delay: float):
        """Used for unit-test mocking."""
        await asyncio.sleep(delay)



from decimal import Decimal

from pydantic import ConfigDict, SecretStr
from pydantic.v1 import Field

from hummingbot.client.config.config_data_types import BaseConnectorConfigMap, ClientFieldData
from hummingbot.connector.derivative.gate_io_perpetual import gate_io_perpetual_constants as CONSTANTS
from hummingbot.core.data_type.trade_fee import TradeFeeSchema

CENTRALIZED = True
EXAMPLE_PAIR = "BTC-USDT"
DEFAULT_FEES = TradeFeeSchema(
    maker_percent_fee_decimal=Decimal("0.00015"),
    taker_percent_fee_decimal=Decimal("0.0005"),
)


class GateIOPerpetualConfigMap(BaseConnectorConfigMap):
    connector: str = Field(default="gate_io_perpetual", client_data=None)
    gate_io_perpetual_api_key: SecretStr = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: f"Enter your {CONSTANTS.EXCHANGE_NAME} API key",
            is_secure=True,
            is_connect_key=True,
            prompt_on_new=True,
        )
    )
    gate_io_perpetual_secret_key: SecretStr = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: f"Enter your {CONSTANTS.EXCHANGE_NAME} secret key",
            is_secure=True,
            is_connect_key=True,
            prompt_on_new=True,
        )
    )
    gate_io_perpetual_user_id: SecretStr = Field(
        default=...,
        client_data=ClientFieldData(
            prompt=lambda cm: f"Enter your {CONSTANTS.EXCHANGE_NAME} user id",
            is_secure=True,
            is_connect_key=True,
            prompt_on_new=True,
        )
    )
    model_config = ConfigDict(title="gate_io_perpetual")


KEYS = GateIOPerpetualConfigMap.construct()



from typing import Any, Dict

from hummingbot.connector.utilities.oms_connector import oms_connector_constants as CONSTANTS


def is_exchange_information_valid(exchange_info: Dict[str, Any]) -> bool:
    """
    Verifies if a trading pair is enabled to operate with based on its exchange information

    :param exchange_info: the exchange information for a trading pair

    :return: True if the trading pair is enabled, False otherwise
    """
    return (
        not exchange_info[CONSTANTS.IS_DISABLED_FIELD]
        and exchange_info[CONSTANTS.SESSION_STATUS_FIELD]
        and "ERR" not in exchange_info[CONSTANTS.SYMBOL_FIELD]
    )



import asyncio
from decimal import Decimal
from typing import Any, Dict, List, Optional, Union

from hummingbot.connector.gateway.amm.gateway_amm_base import GatewayAMMBase
from hummingbot.connector.gateway.gateway_in_flight_order import GatewayInFlightOrder
from hummingbot.core.data_type.common import TradeType
from hummingbot.core.data_type.in_flight_order import OrderState, OrderUpdate
from hummingbot.core.gateway import check_transaction_exceptions
from hummingbot.core.utils.async_utils import safe_ensure_future, safe_gather

s_logger = None
s_decimal_0 = Decimal("0")
s_decimal_NaN = Decimal("nan")


class GatewaySolanaAMM(GatewayAMMBase):
    """
    Defines Solana-specific functions for interacting with AMM protocols via Gateway.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    async def start_network(self):
        if self._trading_required:
            self._status_polling_task = safe_ensure_future(self._status_polling_loop())
            self._get_gas_estimate_task = safe_ensure_future(self.get_gas_estimate())
        self._get_chain_info_task = safe_ensure_future(self.get_chain_info())

    async def stop_network(self):
        if self._status_polling_task is not None:
            self._status_polling_task.cancel()
            self._status_polling_task = None
        if self._get_chain_info_task is not None:
            self._get_chain_info_task.cancel()
            self._get_chain_info_task = None
        if self._get_gas_estimate_task is not None:
            self._get_gas_estimate_task.cancel()
            self._get_chain_info_task = None

    async def _status_polling_loop(self):
        await self.update_balances(on_interval=False)
        while True:
            try:
                self._poll_notifier = asyncio.Event()
                await self._poll_notifier.wait()
                await safe_gather(
                    self.update_balances(on_interval=True),
                    self.update_order_status(self.amm_orders)
                )
                self._last_poll_timestamp = self.current_timestamp
            except asyncio.CancelledError:
                raise
            except Exception as e:
                self.logger().error(str(e), exc_info=True)

    def parse_price_response(
        self,
        base: str,
        quote: str,
        amount: Decimal,
        side: TradeType,
        price_response: Dict[str, Any],
        process_exception: bool = False
    ) -> Optional[Decimal]:
        """
        Parses price response
        :param base: The base asset
        :param quote: The quote asset
        :param amount: amount
        :param side: trade side
        :param price_response: Price response from Gateway.
        :param process_exception: Flag to trigger error on exception
        """
        required_items = ["price", "gasLimit", "gasPrice", "gasCost", "gasPriceToken"]
        if any(item not in price_response.keys() for item in required_items):
            if "info" in price_response.keys():
                self.logger().info(f"Unable to get price. {price_response['info']}")
            else:
                self.logger().info(f"Missing data from price result. Incomplete return result for ({price_response.keys()})")
        else:
            gas_price_token: str = price_response["gasPriceToken"]
            gas_cost: Decimal = Decimal(price_response["gasCost"])
            price: Decimal = Decimal(price_response["price"])
            # self.network_transaction_fee = TokenAmount(gas_price_token, gas_cost)
            if process_exception is True:
                gas_limit: int = int(price_response["gasLimit"])
                exceptions: List[str] = check_transaction_exceptions(
                    balances=self._account_balances,
                    base_asset=base,
                    quote_asset=quote,
                    amount=amount,
                    side=side,
                    gas_limit=gas_limit,
                    gas_cost=gas_cost,
                    gas_asset=gas_price_token,
                    swaps_count=len(price_response.get("swaps", []))
                )
                for index in range(len(exceptions)):
                    self.logger().warning(
                        f"Warning! [{index + 1}/{len(exceptions)}] {side} order - {exceptions[index]}"
                    )
                if len(exceptions) > 0:
                    return None
            return Decimal(str(price))
        return None

    async def update_order_status(self, tracked_orders: List[GatewayInFlightOrder]):
        """
        Calls REST API to get status update for each in-flight amm orders.
        """
        if len(tracked_orders) < 1:
            return

        tx_hash_list: List[str] = await safe_gather(
            *[tracked_order.get_exchange_order_id() for tracked_order in tracked_orders]
        )

        self.logger().info(
            "Polling for order status updates of %d orders. Transaction hashes: %s",
            len(tracked_orders),
            tx_hash_list
        )
        update_results: List[Union[Dict[str, Any], Exception]] = await safe_gather(*[
            self._get_gateway_instance().get_transaction_status(
                self.chain,
                self.network,
                tx_hash
            )
            for tx_hash in tx_hash_list
        ], return_exceptions=True)
        for tracked_order, tx_details in zip(tracked_orders, update_results):
            if isinstance(tx_details, Exception):
                self.logger().error(f"An error occurred fetching transaction status of {tracked_order.client_order_id}")
                continue
            if "txHash" not in tx_details:
                self.logger().error(f"No txHash field for transaction status of {tracked_order.client_order_id}: "
                                    f"{tx_details}.")
                continue
            tx_status: int = tx_details["txStatus"]
            tx_receipt: Optional[Dict[str, Any]] = tx_details["txData"]
            if tx_status == 1 and (tx_receipt is not None):
                # 1: confirmed
                fee: Decimal = Decimal(tx_receipt["meta"]["fee"] / 1000000000)

                self.process_trade_fill_update(tracked_order=tracked_order, fee=fee)

                order_update: OrderUpdate = OrderUpdate(
                    client_order_id=tracked_order.client_order_id,
                    trading_pair=tracked_order.trading_pair,
                    update_timestamp=self.current_timestamp,
                    new_state=OrderState.FILLED,
                )
                self._order_tracker.process_order_update(order_update)
            elif tx_status == 0:
                # 0: fulfilled but not yet confirmed
                pass
            elif tx_status == -1:
                # -1: confirmed but failed
                self.logger().network(
                    f"Error fetching transaction status for the order {tracked_order.client_order_id}: {tx_details}.",
                    app_warning_msg=f"Failed to fetch transaction status for the order {tracked_order.client_order_id}."
                )
                await self._order_tracker.process_order_not_found(tracked_order.client_order_id)



from __future__ import annotations

from decimal import Decimal
from typing import Any, Dict, List, Optional, Union

from pydantic import field_validator
from pydantic.v1 import BaseModel

from hummingbot.core.data_type.common import TradeType
from hummingbot.strategy_v2.executors.position_executor.data_types import TripleBarrierConfig
from hummingbot.strategy_v2.utils.distributions import Distributions


class OrderLevel(BaseModel):
    level: int
    side: TradeType
    order_amount_usd: Decimal
    spread_factor: Decimal = Decimal("0.0")
    order_refresh_time: int = 60
    cooldown_time: int = 0
    triple_barrier_conf: TripleBarrierConfig

    @property
    def level_id(self):
        return f"{self.side.name}_{self.level}"

    @field_validator("order_amount_usd", "spread_factor", mode="before")
    @classmethod
    def float_to_decimal(cls, v):
        return Decimal(v)


class OrderLevelBuilder:
    def __init__(self, n_levels: int):
        """
        Initialize the OrderLevelBuilder with the number of levels.

        Args:
            n_levels (int): The number of order levels.
        """
        self.n_levels = n_levels

    def resolve_input(self, input_data: Union[Decimal | float, List[Decimal | float], Dict[str, Any]]) -> List[Decimal | float | int]:
        """
        Resolve the provided input data into a list of Decimal values.

        Args:
            input_data: The input data to resolve. Can be a single value, list, or dictionary.

        Returns:
            List[Decimal | float | int]: List of resolved Decimal values.
        """
        if isinstance(input_data, Decimal) or isinstance(input_data, float) or isinstance(input_data, int):
            return [input_data] * self.n_levels
        elif isinstance(input_data, list):
            if len(input_data) != self.n_levels:
                raise ValueError(f"List length must match the number of levels: {self.n_levels}")
            return input_data
        elif isinstance(input_data, dict):
            distribution_method = input_data["method"]
            distribution_func = getattr(Distributions, distribution_method, None)
            if not distribution_func:
                raise ValueError(f"Unsupported distribution method: {distribution_method}")
            return distribution_func(self.n_levels, **input_data["params"])
        else:
            raise ValueError(f"Unsupported input data type: {type(input_data)}")

    def build_order_levels(self,
                           amounts: Union[Decimal, List[Decimal], Dict[str, Any]],
                           spreads: Union[Decimal, List[Decimal], Dict[str, Any]],
                           triple_barrier_confs: Union[TripleBarrierConfig, List[TripleBarrierConfig]] = TripleBarrierConfig(),
                           order_refresh_time: Union[int, List[int], Dict[str, Any]] = 60 * 5,
                           cooldown_time: Union[int, List[int], Dict[str, Any]] = 0,
                           sides: Optional[List[TradeType]] = None) -> List[OrderLevel]:
        """
        Build a list of OrderLevels based on the given parameters.

        Args:
            amounts: Amounts to be used for each order level.
            spreads: Spread factors for each order level.
            triple_barrier_confs: Triple barrier configurations.
            order_refresh_time: Time in seconds to wait before refreshing orders.
            cooldown_time: Time in seconds to wait after an order fills before placing a new one.
            sides: Trading sides, either BUY or SELL. Default is both.

        Returns:
            List[OrderLevel]: List of constructed OrderLevel objects.
        """
        if sides is None:
            sides = [TradeType.BUY, TradeType.SELL]

        resolved_amounts = self.resolve_input(amounts)
        resolved_spreads = self.resolve_input(spreads)
        resolved_order_refresh_time = self.resolve_input(order_refresh_time)
        resolved_cooldown_time = self.resolve_input(cooldown_time)

        if not isinstance(triple_barrier_confs, list):
            triple_barrier_confs = [triple_barrier_confs] * self.n_levels

        order_levels = []
        for i in range(self.n_levels):
            for side in sides:
                order_level = OrderLevel(
                    level=i + 1,
                    side=side,
                    order_amount_usd=resolved_amounts[i],
                    spread_factor=resolved_spreads[i],
                    triple_barrier_conf=triple_barrier_confs[i],
                    order_refresh_time=resolved_order_refresh_time[i],
                    cooldown_time=resolved_cooldown_time[i]
                )
                order_levels.append(order_level)

        return order_levels



import asyncio
import logging
from decimal import Decimal
from typing import TYPE_CHECKING, Any, AsyncGenerator, Dict, List, NamedTuple

import hummingbot.connector.exchange.coinbase_advanced_trade.coinbase_advanced_trade_constants as constants
from hummingbot.connector.exchange.coinbase_advanced_trade.coinbase_advanced_trade_web_utils import (
    get_timestamp_from_exchange_time,
)
from hummingbot.core.data_type.common import OrderType, TradeType
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.web_assistant.connections.data_types import WSJSONRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.exchange.coinbase_advanced_trade.coinbase_advanced_trade_exchange import (  # noqa: F401
        CoinbaseAdvancedTradeExchange,
    )


class CoinbaseAdvancedTradeCumulativeUpdate(NamedTuple):
    client_order_id: str
    exchange_order_id: str
    status: str
    trading_pair: str
    fill_timestamp_s: float  # seconds
    average_price: Decimal
    cumulative_base_amount: Decimal
    remainder_base_amount: Decimal
    cumulative_fee: Decimal
    # Needed for tracking existing orders on the exchange
    order_type: OrderType
    trade_type: TradeType
    creation_timestamp_s: float = 0.0  # seconds
    # Coinbase Advanced Trade delivers trade events from the maker's perspective
    is_taker: bool = False


class CoinbaseAdvancedTradeAPIUserStreamDataSource(UserStreamTrackerDataSource):
    """
    UserStreamTrackerDataSource implementation for Coinbase Advanced Trade API.
    """
    _sequence: int = 0
    _logger: HummingbotLogger | logging.Logger | None = None

    @classmethod
    def logger(cls) -> HummingbotLogger | logging.Logger:
        if cls._logger is None:
            name: str = HummingbotLogger.logger_name_for_class(cls)
            cls._logger = logging.getLogger(name)
        return cls._logger

    def __init__(self,
                 auth,
                 trading_pairs: List[str],
                 connector: 'CoinbaseAdvancedTradeExchange',
                 api_factory: WebAssistantsFactory,
                 domain: str = "com"):
        """
        Initialize the CoinbaseAdvancedTradeAPIUserStreamDataSource.

        :param auth: The CoinbaseAdvancedTradeAuth instance for authentication.
        :param trading_pairs: The list of trading pairs to subscribe to.
        :param connector: The CoinbaseAdvancedTradeExchangePairProtocol implementation.
        :param api_factory: The WebAssistantsFactory instance for creating the WSAssistant.
        :param domain: The domain for the WebSocket connection.
        """
        super().__init__()
        self._domain: str = domain
        self._api_factory: WebAssistantsFactory = api_factory
        self._trading_pairs: List[str] = trading_pairs
        self._connector = connector

        self._ws_assistant: WSAssistant | None = None
        self._reset_recv_time: bool = False

    @property
    def last_recv_time(self) -> float:
        """
        Returns the time of the last received message

        :return: the timestamp of the last received message in seconds
        """
        if self._ws_assistant and not self._reset_recv_time:
            return self._ws_assistant.last_recv_time
        return 0

    async def _connected_websocket_assistant(self, pair=None) -> WSAssistant:
        """
        Create and connect the WebSocket assistant.

        :return: The connected WSAssistant instance.
        """
        self._ws_assistant = await self._api_factory.get_ws_assistant()

        await self._ws_assistant.connect(
            ws_url=constants.USER_WSS_URL.format(domain=self._domain),
            ping_timeout=constants.WS_HEARTBEAT_TIME_INTERVAL
        )
        return self._ws_assistant

    async def _subscribe_channels(self, websocket_assistant: WSAssistant) -> None:
        """
        Subscribes to the user events through the provided websocket connection.
        :param websocket_assistant: the websocket assistant used to connect to the exchange
        """
        await self._subscribe_or_unsubscribe(websocket_assistant, constants.WebsocketAction.SUBSCRIBE)

    async def _unsubscribe_channels(self, websocket_assistant: WSAssistant) -> None:
        """
        Unsubscribes to the user events through the provided websocket connection.
        :param websocket_assistant: the websocket assistant used to connect to the exchange
        """
        await self._subscribe_or_unsubscribe(websocket_assistant, constants.WebsocketAction.UNSUBSCRIBE)

    async def _subscribe_or_unsubscribe(
            self,
            websocket_assistant: WSAssistant,
            action: constants.WebsocketAction
    ) -> None:
        """
        Applies the WebsocketAction in argument to the list of channels/pairs through the provided websocket connection.
        :param action: WebsocketAction defined in the CoinbaseAdvancedTradeConstants.
        :param websocket_assistant: the websocket assistant used to connect to the exchange
        """
        """
        https://docs.cdp.coinbase.com/advanced-trade/docs/ws-best-practices

        Recommended to use one subscription per channel
        {
            "type": "subscribe",  # or "unsubscribe"
            "product_ids": [
                "ETH-USD",
                "BTC-USD"   # Possible but not preferred
            ],
            "channel": "user",

            # Complemented by the WSAssistant
            "signature": "XYZ",
            "api_key": "XXX",
            or
            jwt: "JWT",
            "timestamp": 1675974199
        }
        """
        symbols: List[str] = [
            await self._connector.exchange_symbol_associated_to_pair(trading_pair=pair) for pair in self._trading_pairs
        ]

        try:
            for channel in ["heartbeats", constants.WS_USER_SUBSCRIPTION_KEYS]:
                payload = {
                    "type": action.value,
                    "product_ids": symbols,
                    "channel": channel,
                }

                # Change subscription to the channel and pair
                await websocket_assistant.send(WSJSONRequest(payload=payload, is_auth_required=True))
            self.logger().info(
                f"{action.value.capitalize()}-ing to {constants.WS_USER_SUBSCRIPTION_KEYS} for {self._trading_pairs} ...")
        except (asyncio.CancelledError, Exception) as e:
            self.logger().exception(
                f"Unexpected error occurred {action.value.capitalize()}-ing "
                f"to {constants.WS_USER_SUBSCRIPTION_KEYS} for {self._trading_pairs}...\n"
                f"Exception: {e}",
                exc_info=True
            )
            raise

    @staticmethod
    async def _try_except_queue_put(item: Any, queue: asyncio.Queue):
        """
        Try to put the order into the queue, except if the queue is full.
        :param queue: The queue to put the order into.
        """
        try:
            queue.put_nowait(item)
            await asyncio.sleep(0)
        except asyncio.QueueFull:
            try:
                await asyncio.wait_for(queue.put(item), timeout=1.0)
            except asyncio.TimeoutError:
                raise

    async def _process_websocket_messages(self, websocket_assistant: WSAssistant, queue: asyncio.Queue):
        """
        Processes the messages from the websocket connection and puts them into the intermediary queue.
        :param websocket_assistant: the websocket assistant used to connect to the exchange
        :param queue: The intermediary queue to put the messages into.
        """
        async for ws_response in websocket_assistant.iter_messages():  # type: ignore # PyCharm doesn't recognize iter_messages
            data: Dict[str, Any] = ws_response.data

            if 'type' in data and data["type"] == "error":
                if "authentication failure" in data["message"]:
                    self.logger().error(f"authentication error: {data}")
                    await self._subscribe_channels(self._ws_assistant)
                else:
                    self.logger().error(f"Received error message: {data}")
                # Reset last received time
                self._reset_recv_time = True
                return

            self._process_sequence_number(data)

            channel: str = data["channel"]
            if channel == 'user':
                async for order in self._decipher_message(event_message=data):
                    try:
                        # queue.put_nowait(order)
                        await self._try_except_queue_put(item=order, queue=queue)
                    except asyncio.QueueFull:
                        self.logger().exception("Timeout while waiting to put message into raw queue. Message dropped.")
                        raise
            elif channel == 'subscriptions':
                self._process_subscription_message(data)
            elif channel in {"heartbeats"}:
                self._process_heartbeat_message(data)

    def _process_sequence_number(self, data: Dict[str, Any]):
        """
        Processes the sequence number from the websocket message.
        :param data: The message received from the websocket connection.
        """
        if "sequence_num" in data and data["sequence_num"] > self._sequence:
            self.logger().warning(
                f"Received a message with a higher sequence number than the current one. "
                f"Current sequence: {self._sequence}, received sequence: {data['sequence_num']}."
            )

        self._sequence = data["sequence_num"] + 1

    def _process_subscription_message(self, data: Dict[str, Any]):
        """
        Processes the subscription message from the websocket connection.
        :param data: The message received from the websocket connection.
        """
        pass  # self.logger().debug(f"Received subscription message: {data}")

    def _process_heartbeat_message(self, data: Dict[str, Any]):
        """
        Processes the heartbeat message from the websocket connection.
        :param data: The message received from the websocket connection.
        """
        pass  # self.logger().debug(f"Received heartbeat message: {data}")

    async def _decipher_message(self, event_message: Dict[str, Any]) -> AsyncGenerator[Dict[str, Any], None]:
        """
        Streamline the messages for processing by the exchange.
        :param event_message: The message received from the exchange.
        """
        """
        https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#user-channel
        {
          "channel": "user",
          "client_id": "",
          "timestamp": "2023-02-09T20:33:57.609931463Z",
          "sequence_num": 0,
          "events": [
            {
              "type": "snapshot",
              "orders": [
                {
                  "order_id": "XXX",
                  "client_order_id": "YYY",
                  "cumulative_quantity": "0",
                  "leaves_quantity": "0.000994",
                  "avg_price": "0",
                  "total_fees": "0",
                  "status": "OPEN",
                  "product_id": "BTC-USD",
                  "creation_time": "2022-12-07T19:42:18.719312Z",
                  "order_side": "BUY",
                  "order_type": "Limit"
                },
              ]
            }
          ]
        }
        """
        # self.logger().debug(f" '-> _decipher_message: message: {event_message}")

        if not isinstance(event_message["timestamp"], float):
            event_message["timestamp"] = get_timestamp_from_exchange_time(event_message["timestamp"], "s")

        timestamp_s: float = event_message["timestamp"]
        for event in event_message.get("events"):
            for order in event["orders"]:
                try:
                    if order["client_order_id"] != '':
                        order_type: OrderType | None = None
                        if order["order_type"] == "Limit":
                            order_type = OrderType.LIMIT
                        elif order["order_type"] == "Market":
                            order_type = OrderType.MARKET

                        cumulative_order: CoinbaseAdvancedTradeCumulativeUpdate = CoinbaseAdvancedTradeCumulativeUpdate(
                            exchange_order_id=order["order_id"],
                            client_order_id=order["client_order_id"],
                            status=order["status"],
                            trading_pair=await self._connector.trading_pair_associated_to_exchange_symbol(
                                symbol=order["product_id"]
                            ),
                            fill_timestamp_s=timestamp_s,
                            average_price=Decimal(order["avg_price"]),
                            cumulative_base_amount=Decimal(order["cumulative_quantity"]),
                            remainder_base_amount=Decimal(order["leaves_quantity"]),
                            cumulative_fee=Decimal(order["total_fees"]),
                            order_type=order_type,
                            trade_type=TradeType.BUY if order["order_side"] == "BUY" else TradeType.SELL,
                            creation_timestamp_s=get_timestamp_from_exchange_time(order["creation_time"], "s"),
                        )
                        self.logger().debug(f"Yielding order: {cumulative_order}")
                        yield cumulative_order

                except Exception as e:
                    self.logger().exception(f"Failed to create a CumulativeUpdate error {e}\n\t{order}")
                    raise e



from decimal import Decimal
from typing import TYPE_CHECKING

from hummingbot.connector.exchange_base import ExchangeBase
from hummingbot.core.data_type.common import PositionMode

if TYPE_CHECKING:
    from hummingbot.client.config.config_helpers import ClientConfigAdapter

NaN = float("nan")
s_decimal_NaN = Decimal("nan")
s_decimal_0 = Decimal(0)


class DerivativeBase(ExchangeBase):
    """
    DerivativeBase provide extra funtionality in addition to the ExchangeBase for derivative exchanges
    """

    def __init__(self, client_config_map: "ClientConfigAdapter"):
        super().__init__(client_config_map)
        self._funding_info = {}
        self._account_positions = {}
        self._position_mode = None
        self._leverage = {}
        self._funding_payment_span = [0, 0]  # time span(in seconds) before and after funding period when exchanges consider active positions eligible for funding payment

    def set_position_mode(self, position_mode: PositionMode):
        """
        Should set the _position_mode parameter. i.e self._position_mode = position_mode
        This should also be overwritten if the derivative exchange requires interraction to set mode,
        in addition to setting the _position_mode object.
        :param position_mode: ONEWAY or HEDGE position mode
        """
        self._position_mode = position_mode
        return

    def set_leverage(self, trading_pair: str, leverage: int = 1):
        """
        Should set the _leverage parameter. i.e self._leverage = leverage
        This should also be overwritten if the derivative exchange requires interraction to set leverage,
        in addition to setting the _leverage object.
        :param _leverage: leverage to be used
        """
        self._leverage = leverage
        return

    def supported_position_modes(self):
        """
        returns a list containing the modes supported by the derivative
        ONEWAY and/or HEDGE modes
        """
        return [PositionMode.ONEWAY]

    def get_funding_info(self, trading_pair):
        """
        return a dictionary as follows:
        self._trading_info[trading_pair] = {
        "indexPrice": (i.e "21.169488483519444444")
        "markPrice": price used for both pnl on most derivatives (i.e "21.210103847902463671")
        "nextFundingTime": next funding time in unix timestamp (i.e "1612780270")
        "rate": next funding rate as a decimal and not percentage (i.e 0.00007994084744229488)
        }
        """
        raise NotImplementedError



#!/usr/bin/env python
import asyncio
import itertools
import logging
import time
from decimal import Decimal
from functools import lru_cache
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Tuple

import pandas as pd

from hummingbot.client.command.gateway_api_manager import GatewayChainApiManager, begin_placeholder_mode
from hummingbot.client.config.client_config_map import ClientConfigMap
from hummingbot.client.config.config_helpers import (
    ReadOnlyClientConfigAdapter,
    get_connector_class,
    refresh_trade_fees_config,
)
from hummingbot.client.config.security import Security
from hummingbot.client.performance import PerformanceMetrics
from hummingbot.client.settings import AllConnectorSettings, GatewayConnectionSetting, gateway_connector_trading_pairs
from hummingbot.client.ui.completer import load_completer
from hummingbot.client.ui.interface_utils import format_df_for_printout
from hummingbot.core.gateway import get_gateway_paths
from hummingbot.core.gateway.gateway_http_client import GatewayHttpClient
from hummingbot.core.gateway.gateway_status_monitor import GatewayStatus
from hummingbot.core.utils.async_utils import safe_ensure_future, safe_gather
from hummingbot.core.utils.gateway_config_utils import (
    build_config_dict_display,
    build_connector_display,
    build_connector_tokens_display,
    build_list_display,
    build_wallet_display,
    flatten,
    search_configs,
)
from hummingbot.core.utils.ssl_cert import create_self_sign_certs

if TYPE_CHECKING:
    from hummingbot.client.hummingbot_application import HummingbotApplication  # noqa: F401


def ensure_gateway_online(func):
    def wrapper(self, *args, **kwargs):
        if self._gateway_monitor.gateway_status is GatewayStatus.OFFLINE:
            self.logger().error("Gateway is offline")
            return
        return func(self, *args, **kwargs)
    return wrapper


class GatewayCommand(GatewayChainApiManager):
    client_config_map: ClientConfigMap
    _market: Dict[str, Any] = {}

    def __init__(self,  # type: HummingbotApplication
                 client_config_map: ClientConfigMap
                 ):
        super().__init__(client_config_map)
        self.client_config_map = client_config_map

    @ensure_gateway_online
    def gateway_connect(self, connector: str = None):
        safe_ensure_future(self._gateway_connect(connector), loop=self.ev_loop)

    @ensure_gateway_online
    def gateway_status(self):
        safe_ensure_future(self._gateway_status(), loop=self.ev_loop)

    @ensure_gateway_online
    def gateway_balance(self, connector_chain_network: Optional[str] = None):
        if connector_chain_network is not None:
            safe_ensure_future(self._get_balance_for_exchange(
                connector_chain_network), loop=self.ev_loop)
        else:
            safe_ensure_future(self._get_balances(), loop=self.ev_loop)

    @ensure_gateway_online
    def gateway_connector_tokens(self, connector_chain_network: Optional[str], new_tokens: Optional[str]):
        if connector_chain_network is not None and new_tokens is not None:
            safe_ensure_future(self._update_gateway_connector_tokens(
                connector_chain_network, new_tokens), loop=self.ev_loop)
        else:
            safe_ensure_future(self._show_gateway_connector_tokens(
                connector_chain_network), loop=self.ev_loop)

    @ensure_gateway_online
    def gateway_approve_tokens(self, connector_chain_network: Optional[str], tokens: Optional[str]):
        if connector_chain_network is not None and tokens is not None:
            safe_ensure_future(self._update_gateway_approve_tokens(
                connector_chain_network, tokens), loop=self.ev_loop)
        else:
            self.notify(
                "\nPlease specify the connector_chain_network and a token to approve.\n")

    def generate_certs(self):
        safe_ensure_future(self._generate_certs(), loop=self.ev_loop)

    @ensure_gateway_online
    def test_connection(self):
        safe_ensure_future(self._test_connection(), loop=self.ev_loop)

    @ensure_gateway_online
    def gateway_list(self):
        safe_ensure_future(self._gateway_list(), loop=self.ev_loop)

    @ensure_gateway_online
    def gateway_config(self,
                       key: Optional[str] = None,
                       value: str = None):
        if value:
            safe_ensure_future(self._update_gateway_configuration(
                key, value), loop=self.ev_loop)
        else:
            safe_ensure_future(
                self._show_gateway_configuration(key), loop=self.ev_loop)

    async def _test_connection(self):
        # test that the gateway is running
        if await self._get_gateway_instance().ping_gateway():
            self.notify("\nSuccessfully pinged gateway.")
        else:
            self.notify("\nUnable to ping gateway.")

    async def _generate_certs(
            self,       # type: HummingbotApplication
            from_client_password: bool = False,
    ):

        certs_path: str = get_gateway_paths(
            self.client_config_map).local_certs_path.as_posix()

        if not from_client_password:
            with begin_placeholder_mode(self):
                while True:
                    pass_phase = await self.app.prompt(
                        prompt='Enter pass phrase to generate Gateway SSL certifications  >>> ',
                        is_password=True
                    )
                    if pass_phase is not None and len(pass_phase) > 0:
                        break
                    self.notify("Error: Invalid pass phrase")
        else:
            pass_phase = Security.secrets_manager.password.get_secret_value()
        create_self_sign_certs(pass_phase, certs_path)
        self.notify(
            f"Gateway SSL certification files are created in {certs_path}.")
        self._get_gateway_instance().reload_certs(self.client_config_map)

    async def ping_gateway_api(self, max_wait: int) -> bool:
        """
        Try to reach the gateway API for up to max_wait seconds
        """
        now = int(time.time())
        gateway_live = await self._get_gateway_instance().ping_gateway()
        while not gateway_live:
            later = int(time.time())
            if later - now > max_wait:
                return False
            await asyncio.sleep(0.5)
            gateway_live = await self._get_gateway_instance().ping_gateway()
            later = int(time.time())

        return True

    async def _gateway_status(self):
        if self._gateway_monitor.gateway_status is GatewayStatus.ONLINE:
            try:
                status = await self._get_gateway_instance().get_gateway_status()
                if status is None or status == []:
                    self.notify("There are currently no connectors online.")
                else:
                    self.notify(pd.DataFrame(status))
            except Exception:
                self.notify(
                    "\nError: Unable to fetch status of connected Gateway server.")
        else:
            self.notify(
                "\nNo connection to Gateway server exists. Ensure Gateway server is running.")

    async def _update_gateway_configuration(self, key: str, value: Any):
        try:
            response = await self._get_gateway_instance().update_config(key, value)
            self.notify(response["message"])
        except Exception:
            self.notify(
                "\nError: Gateway configuration update failed. See log file for more details.")

    async def _show_gateway_configuration(
        self,  # type: HummingbotApplication
        key: Optional[str] = None,
    ):
        host = self.client_config_map.gateway.gateway_api_host
        port = self.client_config_map.gateway.gateway_api_port
        try:
            config_dict: Dict[str, Any] = await self._gateway_monitor._fetch_gateway_configs()
            if key is not None:
                config_dict = search_configs(config_dict, key)
            self.notify(f"\nGateway Configurations ({host}:{port}):")
            lines = []
            build_config_dict_display(lines, config_dict)
            self.notify("\n".join(lines))

        except asyncio.CancelledError:
            raise
        except Exception:
            remote_host = ':'.join([host, port])
            self.notify(f"\nError: Connection to Gateway {remote_host} failed")

    async def _gateway_connect(
            self,           # type: HummingbotApplication
            connector: str = None
    ):
        with begin_placeholder_mode(self):
            gateway_connections_conf: List[Dict[str,
                                                str]] = GatewayConnectionSetting.load()
            if connector is None:
                if len(gateway_connections_conf) < 1:
                    self.notify("No existing connection.\n")
                else:
                    connector_df: pd.DataFrame = build_connector_display(
                        gateway_connections_conf)
                    self.notify(connector_df.to_string(index=False))
            else:
                # get available networks
                connector_configs: Dict[str, Any] = await self._get_gateway_instance().get_connectors()
                connector_config: List[Dict[str, Any]] = [
                    d for d in connector_configs["connectors"] if d["name"] == connector
                ]
                if len(connector_config) < 1:
                    self.notify(
                        f"No available blockchain networks available for the connector '{connector}'.")
                    return
                available_networks: List[Dict[str, Any]
                                         ] = connector_config[0]["available_networks"]
                trading_type: str = connector_config[0]["trading_type"][0]
                additional_prompts: Dict[str, str] = connector_config[0].get(  # These will be stored locally.
                    # If Gateway requires additional, prompts with secure info,
                    "additional_add_wallet_prompts",
                    # a new attribute must be added (e.g. additional_secure_add_wallet_prompts)
                    {}
                )

                # ask user to select a chain. Automatically select if there is only one.
                chains: List[str] = [d['chain'] for d in available_networks]
                chain: str

                if len(chains) == 1:
                    chain = chains[0]
                    self.notify(f"Only {chain} chain is available. Selected automatically.")
                else:
                    # chains as options
                    while True:
                        self.app.input_field.completer.set_gateway_chains(chains)
                        chain = await self.app.prompt(
                            prompt=f"Which chain do you want {connector} to connect to? ({', '.join(chains)}) >>> "
                        )
                        if self.app.to_stop_config:
                            self.app.to_stop_config = False
                            return

                        if chain in chains:
                            break
                        self.notify(f"{chain} chain not supported.\n")

                # ask user to select a network. Automatically select if there is only one.
                networks: List[str] = list(
                    itertools.chain.from_iterable(
                        [d['networks'] for d in available_networks if d['chain'] == chain])
                )

                network: str
                if len(networks) == 1:
                    network = networks[0]
                    self.notify(f"Only {network} network is available. Selected automatically.")
                else:
                    while True:
                        self.app.input_field.completer.set_gateway_networks(networks)
                        network = await self.app.prompt(
                            prompt=f"Which network do you want {connector} to connect to? ({', '.join(networks)}) >>> "
                        )
                        if self.app.to_stop_config:
                            return
                        if network in networks:
                            break
                        self.notify("Error: Invalid network")

                # test you can connect to the uri, otherwise request the url
                await self._test_node_url_from_gateway_config(chain, network, attempt_connection=False)

                if self.app.to_stop_config:
                    return

                # get wallets for the selected chain
                wallets_response: List[Dict[str, Any]] = await self._get_gateway_instance().get_wallets()
                matching_wallets: List[Dict[str, Any]] = [
                    w for w in wallets_response if w["chain"] == chain]
                wallets: List[str]
                if len(matching_wallets) < 1:
                    wallets = []
                else:
                    wallets = matching_wallets[0]['walletAddresses']

                # if the user has no wallet, ask them to select one
                if len(wallets) < 1 or len(additional_prompts) != 0:
                    wallet_address, additional_prompt_values = await self._prompt_for_wallet_address(
                        chain=chain, network=network, additional_prompts=additional_prompts
                    )

                # the user has a wallet. Ask if they want to use it or create a new one.
                else:
                    # print table
                    while True:
                        use_existing_wallet: str = await self.app.prompt(
                            prompt=f"Do you want to connect to {chain}-{network} with one of your existing wallets on "
                                   f"Gateway? (Yes/No) >>> "
                        )
                        if self.app.to_stop_config:
                            return
                        if use_existing_wallet in ["Y", "y", "Yes", "yes", "N", "n", "No", "no"]:
                            break
                        self.notify(
                            "Invalid input. Please try again or exit config [CTRL + x].\n")

                    self.app.clear_input()
                    # they use an existing wallet
                    if use_existing_wallet is not None and use_existing_wallet in ["Y", "y", "Yes", "yes"]:
                        native_token: str = await self._get_native_currency_symbol(chain, network)
                        wallet_table: List[Dict[str, Any]] = []
                        for w in wallets:
                            balances: Dict[str, Any] = await self._get_gateway_instance().get_balances(
                                chain, network, w, [native_token], connector
                            )
                            balance = (
                                balances['balances'].get(native_token)
                                or balances['balances']['total'].get(native_token)
                            )
                            wallet_table.append(
                                {"balance": balance, "address": w})

                        wallet_df: pd.DataFrame = build_wallet_display(
                            native_token, wallet_table)
                        self.notify(wallet_df.to_string(index=False))
                        self.app.input_field.completer.set_list_gateway_wallets_parameters(
                            wallets_response, chain)
                        additional_prompt_values = {}

                        while True:
                            wallet_address: str = await self.app.prompt(prompt="Select a gateway wallet >>> ")
                            if self.app.to_stop_config:
                                return
                            if wallet_address in wallets:
                                self.notify(
                                    f"You have selected {wallet_address}.")
                                break
                            self.notify("Error: Invalid wallet address")

                    # they want to create a new wallet even though they have other ones
                    else:
                        while True:
                            try:
                                wallet_address, additional_prompt_values = await self._prompt_for_wallet_address(
                                    chain=chain, network=network, additional_prompts=additional_prompts
                                )
                                break
                            except Exception:
                                self.notify(
                                    "Error adding wallet. Check private key.\n")

                        # display wallet balance
                        native_token: str = await self._get_native_currency_symbol(chain, network)
                        balances: Dict[str, Any] = await self._get_gateway_instance().get_balances(
                            chain, network, wallet_address, [
                                native_token], connector
                        )
                        wallet_table: List[Dict[str, Any]] = [{"balance": balances['balances'].get(
                            native_token) or balances['balances']['total'].get(native_token), "address": wallet_address}]
                        wallet_df: pd.DataFrame = build_wallet_display(
                            native_token, wallet_table)
                        self.notify(wallet_df.to_string(index=False))

                self.app.clear_input()

                # write wallets to Gateway connectors settings.
                GatewayConnectionSetting.upsert_connector_spec(
                    connector_name=connector,
                    chain=chain,
                    network=network,
                    trading_type=trading_type,
                    wallet_address=wallet_address,
                    additional_prompt_values=additional_prompt_values,
                )
                self.notify(
                    f"The {connector} connector now uses wallet {wallet_address} on {chain}-{network}")

                # update AllConnectorSettings and fee overrides.
                AllConnectorSettings.create_connector_settings()
                AllConnectorSettings.initialize_paper_trade_settings(
                    self.client_config_map.paper_trade.paper_trade_exchanges
                )
                await refresh_trade_fees_config(self.client_config_map)

                # Reload completer here to include newly added gateway connectors
                self.app.input_field.completer = load_completer(self)

    async def _prompt_for_wallet_address(
        self,           # type: HummingbotApplication
        chain: str,
        network: str,
        additional_prompts: Dict[str, str],
    ) -> Tuple[Optional[str], Dict[str, str]]:
        self.app.clear_input()
        self.placeholder_mode = True
        wallet_private_key = await self.app.prompt(
            prompt=f"Enter your {chain}-{network} wallet private key >>> ",
            is_password=True
        )
        self.app.clear_input()
        if self.app.to_stop_config:
            return

        additional_prompt_values = {}

        for field, prompt in additional_prompts.items():
            value = await self.app.prompt(prompt=prompt, is_password=True)
            self.app.clear_input()
            if self.app.to_stop_config:
                return
            additional_prompt_values[field] = value

        response: Dict[str, Any] = await self._get_gateway_instance().add_wallet(
            chain, network, wallet_private_key, **additional_prompt_values
        )
        wallet_address: str = response["address"]
        return wallet_address, additional_prompt_values

    async def _get_balance_for_exchange(self, exchange_name: str):
        gateway_connections = GatewayConnectionSetting.load()
        gateway_instance = GatewayHttpClient.get_instance(self.client_config_map)
        network_timeout = float(self.client_config_map.commands_timeout.other_commands_timeout)
        self.notify("Updating gateway balances, please wait...")
        conf: Optional[Dict[str, str]] = GatewayConnectionSetting.get_connector_spec_from_market_name(
            exchange_name)
        if conf is None:
            self.notify(
                f"'{exchange_name}' is not available. You can add and review exchange with 'gateway connect'.")
        else:
            chain, network, address = (
                conf["chain"], conf["network"], conf["wallet_address"]
            )
            # Add native token to the tokens list
            native_token = await self._get_native_currency_symbol(chain, network)
            tokens_str = conf.get("tokens", "")
            tokens = [token.strip() for token in tokens_str.split(',')] if tokens_str else []
            if native_token not in tokens:
                tokens.append(native_token)

            connector_chain_network = [
                w for w in gateway_connections
                if w["chain"] == chain and
                w["network"] == network and
                w["connector"] == conf["connector"] and
                w["trading_type"] == conf["trading_type"]
            ]

            connector = connector_chain_network[0]['connector']
            exchange_key = f"{connector}_{chain}_{network}"

            try:
                single_ex_bal = await asyncio.wait_for(
                    self.single_balance_exc(exchange_name, self.client_config_map), network_timeout
                )

                allowance_resp = await gateway_instance.get_allowances(
                    chain, network, address, tokens, connector_chain_network[0]["connector"], fail_silently=True
                )

                rows = []
                for exchange, bals in single_ex_bal.items():
                    if exchange_key == exchange:
                        rows = []
                        for token, bal in bals.items():
                            # Handle errors in allowance_responses_list
                            if allowance_resp.get("approvals") is None:
                                allowance = Decimal("0")
                            else:
                                allowance = allowance_resp["approvals"].get(token, Decimal("0"))

                            rows.append({
                                "Symbol": token.upper(),
                                "Balance": PerformanceMetrics.smart_round(Decimal(str(bal)), 4),
                                "Allowance": PerformanceMetrics.smart_round(Decimal(str(allowance)), 4) if float(allowance) < 999999 else "999999+",

                            })

                df = pd.DataFrame(data=rows, columns=["Symbol", "Balance", "Allowance"])
                df.sort_values(by=["Symbol"], inplace=True)

                self.notify(f"\nConnector: {exchange_key}")
                self.notify(f"Wallet_Address: {address}")

                if df.empty:
                    self.notify("You have no balance on this exchange.")
                else:
                    lines = [
                        "    " + line for line in df.to_string(index=False).split("\n")
                    ]
                    self.notify("\n".join(lines))

            except asyncio.TimeoutError:
                self.notify("\nA network error prevented the balances from updating. See logs for more details.")
                raise

    async def _get_balances(self):
        network_connections = GatewayConnectionSetting.load()
        gateway_instance = GatewayHttpClient.get_instance(self.client_config_map)
        network_timeout = float(self.client_config_map.commands_timeout.other_commands_timeout)
        self.notify("Updating gateway balances, please wait...")

        try:
            bal_resp = await asyncio.wait_for(
                self.all_balances_all_exc(self.client_config_map), network_timeout
            )
            allowance_tasks = []

            for conf in network_connections:
                chain, network, address = (
                    conf["chain"], conf["network"], conf["wallet_address"]
                )

                # Add native token to the tokens list
                native_token = await self._get_native_currency_symbol(chain, network)
                tokens_str = conf.get("tokens", "")
                tokens = [token.strip() for token in tokens_str.split(',')] if tokens_str else []
                if native_token not in tokens:
                    tokens.append(native_token)

                connector_chain_network = [
                    w for w in network_connections
                    if w["chain"] == chain and
                    w["network"] == network and
                    w["connector"] == conf["connector"] and
                    w["trading_type"] == conf["trading_type"]
                ]

                allowance_resp = gateway_instance.get_allowances(
                    chain, network, address, tokens, connector_chain_network[0]["connector"], fail_silently=True
                )
                allowance_tasks.append(allowance_resp)

            # Gather balances and allowance responses asynchronously
            allowance_responses_list = await asyncio.gather(*allowance_tasks)

            for conf in network_connections:
                index = network_connections.index(conf)
                chain, network, address, connector = conf["chain"], conf["network"], conf["wallet_address"], conf["connector"]
                exchange_key = f'{connector}_{chain}_{network}'
                exchange_found = False
                for exchange, bals in bal_resp.items():
                    if exchange_key == exchange:
                        exchange_found = True
                        rows = []
                        for token, bal in bals.items():
                            # Handle errors in allowance_responses_list
                            if allowance_responses_list[index].get("approvals") is None:
                                allowance = Decimal("0")
                            else:
                                allowance = allowance_responses_list[index]["approvals"].get(token, Decimal("0"))

                            rows.append({
                                "Symbol": token.upper(),
                                "Balance": PerformanceMetrics.smart_round(Decimal(str(bal)), 4),
                                "Allowance": PerformanceMetrics.smart_round(Decimal(str(allowance)), 4) if float(allowance) < 999999 else "999999+",

                            })

                        df = pd.DataFrame(data=rows, columns=["Symbol", "Balance", "Allowance"])
                        df.sort_values(by=["Symbol"], inplace=True)

                        self.notify(f"\nConnector: {exchange_key}")
                        self.notify(f"Wallet_Address: {address}")

                        if df.empty:
                            self.notify("You have no balance and allowances on this exchange.")
                        else:
                            lines = [
                                "    " + line for line in df.to_string(index=False).split("\n")
                            ]
                            self.notify("\n".join(lines))
                if not exchange_found:
                    self.notify(f"\nConnector: {exchange_key}")
                    self.notify(f"Wallet_Address: {address}")
                    self.notify("You have no balance and allowances on this exchange.")

        except asyncio.TimeoutError:
            self.notify("\nA network error prevented the balances and allowances from updating. See logs for more details.")
            raise

    def connect_markets(exchange, client_config_map: ClientConfigMap, **api_details):
        connector = None
        conn_setting = AllConnectorSettings.get_connector_settings()[exchange]
        if api_details or conn_setting.uses_gateway_generic_connector():
            connector_class = get_connector_class(exchange)
            read_only_client_config = ReadOnlyClientConfigAdapter.lock_config(
                client_config_map)
            init_params = conn_setting.conn_init_parameters(
                trading_pairs=gateway_connector_trading_pairs(
                    conn_setting.name),
                api_keys=api_details,
                client_config_map=read_only_client_config,
            )

            # collect trading pairs from the gateway connector settings
            trading_pairs: List[str] = gateway_connector_trading_pairs(
                conn_setting.name)

            # collect unique trading pairs that are for balance reporting only
            if conn_setting.uses_gateway_generic_connector():
                config: Optional[Dict[str, str]] = GatewayConnectionSetting.get_connector_spec_from_market_name(
                    conn_setting.name)
                if config is not None:
                    existing_pairs = set(
                        flatten([x.split("-") for x in trading_pairs]))

                    other_tokens: Set[str] = set(
                        config.get("tokens", "").split(","))
                    other_tokens.discard("")
                    tokens: List[str] = [
                        t for t in other_tokens if t not in existing_pairs]
                    if tokens != [""]:
                        trading_pairs.append("-".join(tokens))

            connector = connector_class(**init_params)
        return connector

    @staticmethod
    async def _update_balances(market) -> Optional[str]:
        try:
            await market._update_balances()
        except Exception as e:
            logging.getLogger().debug(
                f"Failed to update balances for {market}", exc_info=True)
            return str(e)
        return None

    async def add_gateway_exchange(self, exchange, client_config_map: ClientConfigMap, **api_details) -> Optional[str]:
        self._market.pop(exchange, None)
        is_gateway_markets = self.is_gateway_markets(exchange)
        if is_gateway_markets:
            market = GatewayCommand.connect_markets(
                exchange, client_config_map, **api_details)
            if not market:
                return "API keys have not been added."
            err_msg = await GatewayCommand._update_balances(market)
            if err_msg is None:
                self._market[exchange] = market
            return err_msg

    def all_balance(self, exchange) -> Dict[str, Decimal]:
        if exchange not in self._market:
            return {}
        return self._market[exchange].get_all_balances()

    async def update_exchange_balances(self, exchange_name: str, client_config_map: ClientConfigMap) -> Optional[Tuple[Dict[str, Any], Dict[str, Any]]]:
        is_gateway_markets = self.is_gateway_markets(exchange_name)
        if is_gateway_markets and exchange_name in self._market:
            del self._market[exchange_name]
        if exchange_name in self._market:
            return await self._update_balances(self._market[exchange_name])
        else:
            await Security.wait_til_decryption_done()
            api_keys = Security.api_keys(
                exchange_name) if not is_gateway_markets else {}
            return await self.add_gateway_exchange(exchange_name, client_config_map, **api_keys)

    @staticmethod
    @lru_cache(maxsize=10)
    def is_gateway_markets(exchange_name: str) -> bool:
        return (
            exchange_name in sorted(
                AllConnectorSettings.get_gateway_amm_connector_names()
            )
        )

    async def update_exchange(
        self,
        client_config_map: ClientConfigMap,
        reconnect: bool = False,
        exchanges: Optional[List[str]] = None
    ) -> Dict[str, Optional[str]]:
        exchanges = exchanges or []
        tasks = []
        # Update user balances
        if len(exchanges) == 0:
            exchanges = [
                cs.name for cs in AllConnectorSettings.get_connector_settings().values()]
        exchanges: List[str] = [
            cs.name
            for cs in AllConnectorSettings.get_connector_settings().values()
            if not cs.use_ethereum_wallet
            and cs.name in exchanges
            and not cs.name.endswith("paper_trade")
        ]

        if reconnect:
            self._market.clear()
        for exchange in exchanges:
            tasks.append(self.update_exchange_balances(
                exchange, client_config_map))
        results = await safe_gather(*tasks)
        return {ex: err_msg for ex, err_msg in zip(exchanges, results)}

    async def all_balances_all_exc(self, client_config_map: ClientConfigMap) -> Dict[str, Dict[str, Decimal]]:
        # Waits for the update_exchange method to complete with the provided client_config_map
        await self.update_exchange(client_config_map)
        return {k: v.get_all_balances() for k, v in sorted(self._market.items(), key=lambda x: x[0])}

    async def balance(self, exchange, client_config_map: ClientConfigMap, *symbols) -> Dict[str, Decimal]:
        if await self.update_exchange_balances(exchange, client_config_map) is None:
            results = {}
            for token, bal in self.all_balance(exchange).items():
                matches = [s for s in symbols if s.lower() == token.lower()]
                if matches:
                    results[matches[0]] = bal
            return results

    async def update_exch(
        self,
        exchange: str,
        client_config_map: ClientConfigMap,
        reconnect: bool = False,
        exchanges: Optional[List[str]] = None
    ) -> Dict[str, Optional[str]]:
        exchanges = exchanges or []
        tasks = []
        if reconnect:
            self._market.clear()
        tasks.append(self.update_exchange_balances(exchange, client_config_map))
        results = await safe_gather(*tasks)
        return {ex: err_msg for ex, err_msg in zip(exchanges, results)}

    async def single_balance_exc(self, exchange, client_config_map: ClientConfigMap) -> Dict[str, Dict[str, Decimal]]:
        # Waits for the update_exchange method to complete with the provided client_config_map
        await self.update_exch(exchange, client_config_map)
        return {k: v.get_all_balances() for k, v in sorted(self._market.items(), key=lambda x: x[0])}

    async def _show_gateway_connector_tokens(
            self,           # type: HummingbotApplication
            connector_chain_network: str = None
    ):
        """
        Display connector tokens that hummingbot will report balances for
        """
        if connector_chain_network is None:
            gateway_connections_conf: Dict[str, List[str]] = GatewayConnectionSetting.load()
            if len(gateway_connections_conf) < 1:
                self.notify("No existing connection.\n")
            else:
                connector_df: pd.DataFrame = build_connector_tokens_display(gateway_connections_conf)
                self.notify(connector_df.to_string(index=False))
        else:
            conf: Optional[Dict[str, List[str]]] = GatewayConnectionSetting.get_connector_spec_from_market_name(connector_chain_network)
            if conf is not None:
                connector_df: pd.DataFrame = build_connector_tokens_display([conf])
                self.notify(connector_df.to_string(index=False))
            else:
                self.notify(
                    f"There is no gateway connection for {connector_chain_network}.\n")

    async def _update_gateway_connector_tokens(
            self,           # type: HummingbotApplication
            connector_chain_network: str,
            new_tokens: str,
    ):
        """
        Allow the user to input tokens whose balances they want to monitor are.
        These are not tied to a strategy, rather to the connector-chain-network
        tuple. This has no influence on what tokens the user can use with a
        connector-chain-network and a particular strategy. This is only for
        report balances.
        """
        conf: Optional[Dict[str, str]] = GatewayConnectionSetting.get_connector_spec_from_market_name(
            connector_chain_network)

        if conf is None:
            self.notify(
                f"'{connector_chain_network}' is not available. You can add and review available gateway connectors with the command 'gateway connect'.")
        else:
            GatewayConnectionSetting.upsert_connector_spec_tokens(connector_chain_network, new_tokens)
            self.notify(
                f"The 'gateway balance' command will now report token balances {new_tokens} for '{connector_chain_network}'.")

    async def _gateway_list(
        self           # type: HummingbotApplication
    ):
        connector_list: List[Dict[str, Any]] = await self._get_gateway_instance().get_connectors()
        connectors_tiers: List[Dict[str, Any]] = []
        for connector in connector_list["connectors"]:
            available_networks: List[Dict[str, Any]
                                     ] = connector["available_networks"]
            chains: List[str] = [d['chain'] for d in available_networks]
            connector['chains'] = chains
            connectors_tiers.append(connector)
        connectors_df: pd.DataFrame = build_list_display(connectors_tiers)
        lines = ["    " + line for line in format_df_for_printout(
            connectors_df,
            table_format=self.client_config_map.tables_format).split("\n")]
        self.notify("\n".join(lines))

    async def _update_gateway_approve_tokens(
            self,           # type: HummingbotApplication
            connector_chain_network: str,
            tokens: str,
    ):
        """
        Allow the user to approve tokens for spending.
        """
        # get connector specs
        conf: Optional[Dict[str, str]] = GatewayConnectionSetting.get_connector_spec_from_market_name(
            connector_chain_network)
        if conf is None:
            self.notify(
                f"'{connector_chain_network}' is not available. You can add and review available gateway connectors with the command 'gateway connect'.")
        else:
            self.logger().info(
                f"Connector {conf['connector']} Tokens {tokens} will now be approved for spending for '{connector_chain_network}'.")
            # get wallets for the selected chain
            gateway_connections_conf: List[Dict[str,
                                                str]] = GatewayConnectionSetting.load()
            if len(gateway_connections_conf) < 1:
                self.notify("No existing wallet.\n")
                return
            connector_wallet: List[Dict[str, Any]] = [w for w in gateway_connections_conf if w["chain"] ==
                                                      conf['chain'] and w["connector"] == conf['connector'] and w["network"] == conf['network']]
            try:
                resp: Dict[str, Any] = await self._get_gateway_instance().approve_token(conf['chain'], conf['network'], connector_wallet[0]['wallet_address'], tokens, conf['connector'])
                transaction_hash: Optional[str] = resp.get(
                    "approval", {}).get("hash")
                displayed_pending: bool = False
                while True:
                    pollResp: Dict[str, Any] = await self._get_gateway_instance().get_transaction_status(conf['chain'], conf['network'], transaction_hash)
                    transaction_status: Optional[str] = pollResp.get(
                        "txStatus")
                    if transaction_status == 1:
                        self.logger().info(
                            f"Token {tokens} is approved for spending for '{conf['connector']}' for Wallet: {connector_wallet[0]['wallet_address']}.")
                        self.notify(
                            f"Token {tokens} is approved for spending for '{conf['connector']}' for Wallet: {connector_wallet[0]['wallet_address']}.")
                        break
                    elif transaction_status == 2:
                        if not displayed_pending:
                            self.logger().info(
                                f"Token {tokens} approval transaction is pending. Transaction hash: {transaction_hash}")
                            displayed_pending = True
                            await asyncio.sleep(2)
                        continue
                    else:
                        self.logger().info(
                            f"Tokens {tokens} is not approved for spending. Please use manual approval.")
                        self.notify(
                            f"Tokens {tokens} is not approved for spending. Please use manual approval.")
                        break

            except Exception as e:
                self.logger().error(f"Error approving tokens: {e}")
                return

    def _get_gateway_instance(
        self  # type: HummingbotApplication
    ) -> GatewayHttpClient:
        gateway_instance = GatewayHttpClient.get_instance(
            self.client_config_map)
        return gateway_instance



#!/usr/bin/env python

import asyncio
import aiohttp
import base64
import logging
import pandas as pd
from typing import (
    Dict,
    Optional,
    Tuple,
    AsyncIterable
)
import pickle
import time
import websockets
from websockets.exceptions import ConnectionClosed

import conf
from hummingbot.connector.exchange.binance.binance_order_book import BinanceOrderBook
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.logger import HummingbotLogger
from hummingbot.core.data_type.order_book_tracker_entry import OrderBookTrackerEntry


class RemoteAPIOrderBookDataSource(OrderBookTrackerDataSource):
    SNAPSHOT_REST_URL = "https://api.coinalpha.com/order_book_tracker/snapshot"
    SNAPSHOT_STREAM_URL = "wss://api.coinalpha.com/ws/order_book_tracker/snapshot_stream"
    DIFF_STREAM_URL = "wss://api.coinalpha.com/ws/order_book_tracker/diff_stream"

    MESSAGE_TIMEOUT = 30.0
    PING_TIMEOUT = 10.0

    _raobds_logger: Optional[HummingbotLogger] = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._raobds_logger is None:
            cls._raobds_logger = logging.getLogger(__name__)
        return cls._raobds_logger

    def __init__(self):
        super().__init__()
        self._client_session: Optional[aiohttp.ClientSession] = None

    @property
    def authentication_headers(self) -> Dict[str, str]:
        auth_str: str = f"{conf.coinalpha_order_book_api_username}:{conf.coinalpha_order_book_api_password}"
        encoded_auth: str = base64.standard_b64encode(auth_str.encode("utf8")).decode("utf8")
        return {
            "Authorization": f"Basic {encoded_auth}"
        }

    async def get_client_session(self) -> aiohttp.ClientSession:
        if self._client_session is None:
            self._client_session = aiohttp.ClientSession()
        return self._client_session

    async def get_tracking_pairs(self) -> Dict[str, OrderBookTrackerEntry]:
        auth: aiohttp.BasicAuth = aiohttp.BasicAuth(login=conf.coinalpha_order_book_api_username,
                                                    password=conf.coinalpha_order_book_api_password)
        client_session: aiohttp.ClientSession = await self.get_client_session()
        response: aiohttp.ClientResponse = await client_session.get(self.SNAPSHOT_REST_URL, auth=auth)
        timestamp: float = time.time()
        if response.status != 200:
            raise EnvironmentError(f"Error fetching order book tracker snapshot from {self.SNAPSHOT_REST_URL}.")

        binary_data: bytes = await response.read()
        order_book_tracker_data: Dict[str, Tuple[pd.DataFrame, pd.DataFrame]] = pickle.loads(binary_data)
        retval: Dict[str, OrderBookTrackerEntry] = {}

        for trading_pair, (bids_df, asks_df) in order_book_tracker_data.items():
            order_book: BinanceOrderBook = BinanceOrderBook()
            order_book.apply_numpy_snapshot(bids_df.values, asks_df.values)
            retval[trading_pair] = OrderBookTrackerEntry(trading_pair, timestamp, order_book)

        return retval

    async def _inner_messages(self,
                              ws: websockets.WebSocketClientProtocol) -> AsyncIterable[str]:
        # Terminate the recv() loop as soon as the next message timed out, so the outer loop can reconnect.
        try:
            while True:
                try:
                    msg: str = await asyncio.wait_for(ws.recv(), timeout=self.MESSAGE_TIMEOUT)
                    yield msg
                except asyncio.TimeoutError:
                    pong_waiter = await ws.ping()
                    await asyncio.wait_for(pong_waiter, timeout=self.PING_TIMEOUT)
        except asyncio.TimeoutError:
            self.logger().warning("WebSocket ping timed out. Going to reconnect...")
            return
        except ConnectionClosed:
            return
        finally:
            await ws.close()

    async def listen_for_order_book_diffs(self, ev_loop: asyncio.BaseEventLoop, output: asyncio.Queue):
        while True:
            try:
                async with websockets.connect(self.DIFF_STREAM_URL,
                                              extra_headers=self.authentication_headers) as ws:
                    ws: websockets.WebSocketClientProtocol = ws
                    async for msg in self._inner_messages(ws):
                        output.put_nowait(msg)
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().error("Unexpected error with WebSocket connection. Retrying after 30 seconds...",
                                    exc_info=True)
                await asyncio.sleep(30.0)

    async def listen_for_order_book_snapshots(self, ev_loop: asyncio.BaseEventLoop, output: asyncio.Queue):
        while True:
            try:
                async with websockets.connect(self.SNAPSHOT_STREAM_URL,
                                              extra_headers=self.authentication_headers) as ws:
                    ws: websockets.WebSocketClientProtocol = ws
                    async for msg in self._inner_messages(ws):
                        output.put_nowait(msg)
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().error("Unexpected error with WebSocket connection. Retrying after 30 seconds...",
                                    exc_info=True)
                await asyncio.sleep(30.0)



import logging
from decimal import Decimal
from typing import Dict

import pandas as pd

from hummingbot.client.config.client_config_map import ClientConfigMap
from hummingbot.client.config.config_helpers import ClientConfigAdapter, get_connector_class
from hummingbot.client.settings import AllConnectorSettings, ConnectorType
from hummingbot.connector.connector_base import ConnectorBase
from hummingbot.core.data_type.common import PriceType
from hummingbot.data_feed.candles_feed.candles_base import CandlesBase
from hummingbot.data_feed.candles_feed.candles_factory import CandlesFactory
from hummingbot.data_feed.candles_feed.data_types import CandlesConfig, HistoricalCandlesConfig
from hummingbot.data_feed.market_data_provider import MarketDataProvider

# Set up logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class BacktestingDataProvider(MarketDataProvider):
    CONNECTOR_TYPES = [ConnectorType.CLOB_SPOT, ConnectorType.CLOB_PERP, ConnectorType.Exchange,
                       ConnectorType.Derivative]
    EXCLUDED_CONNECTORS = ["hyperliquid_perpetual", "dydx_perpetual", "cube", "vertex",
                           "coinbase_advanced_trade", "kraken", "dydx_v4_perpetual", "hitbtc",
                           "hyperliquid", "injective_v2_perpetual", "injective_v2"]

    def __init__(self, connectors: Dict[str, ConnectorBase]):
        super().__init__(connectors)
        self.start_time = None
        self.end_time = None
        self.prices = {}
        self._time = None
        self.trading_rules = {}
        self.conn_settings = AllConnectorSettings.get_connector_settings()
        self.connectors = {name: self.get_connector(name) for name, settings in self.conn_settings.items()
                           if settings.type in self.CONNECTOR_TYPES and name not in self.EXCLUDED_CONNECTORS and
                           "testnet" not in name}

    def get_connector(self, connector_name: str):
        conn_setting = self.conn_settings.get(connector_name)
        if conn_setting is None:
            logger.error(f"Connector {connector_name} not found")
            raise ValueError(f"Connector {connector_name} not found")

        client_config_map = ClientConfigAdapter(ClientConfigMap())
        init_params = conn_setting.conn_init_parameters(
            trading_pairs=[],
            trading_required=False,
            api_keys=self.get_connector_config_map(connector_name),
            client_config_map=client_config_map,
        )
        connector_class = get_connector_class(connector_name)
        connector = connector_class(**init_params)
        return connector

    @staticmethod
    def get_connector_config_map(connector_name: str):
        connector_config = AllConnectorSettings.get_connector_config_keys(connector_name)
        return {key: "" for key in connector_config.__fields__.keys() if key != "connector"}

    def get_trading_rules(self, connector_name: str, trading_pair: str):
        """
        Retrieves the trading rules from the specified connector.
        :param connector_name: str
        :return: Trading rules.
        """
        return self.trading_rules[connector_name][trading_pair]

    def time(self):
        return self._time

    async def initialize_trading_rules(self, connector_name: str):
        if len(self.trading_rules.get(connector_name, {})) == 0:
            connector = self.connectors.get(connector_name)
            await connector._update_trading_rules()
            self.trading_rules[connector_name] = connector.trading_rules

    async def initialize_candles_feed(self, config: CandlesConfig):
        await self.get_candles_feed(config)

    def update_backtesting_time(self, start_time: int, end_time: int):
        self.start_time = start_time
        self.end_time = end_time
        self._time = start_time

    async def get_candles_feed(self, config: CandlesConfig):
        """
        Retrieves or creates and starts a candle feed based on the given configuration.
        If an existing feed has a higher or equal max_records, it is reused.
        :param config: CandlesConfig
        :return: Candle feed instance.
        """
        key = self._generate_candle_feed_key(config)
        existing_feed = self.candles_feeds.get(key, pd.DataFrame())

        if not existing_feed.empty:
            existing_feed_start_time = existing_feed["timestamp"].min()
            existing_feed_end_time = existing_feed["timestamp"].max()
            if existing_feed_start_time <= self.start_time and existing_feed_end_time >= self.end_time:
                return existing_feed
        # Create a new feed or restart the existing one with updated max_records
        candle_feed = CandlesFactory.get_candle(config)
        candles_buffer = config.max_records * CandlesBase.interval_to_seconds[config.interval]
        candles_df = await candle_feed.get_historical_candles(config=HistoricalCandlesConfig(
            connector_name=config.connector,
            trading_pair=config.trading_pair,
            interval=config.interval,
            start_time=self.start_time - candles_buffer,
            end_time=self.end_time,
        ))
        self.candles_feeds[key] = candles_df
        return candles_df

    def get_candles_df(self, connector_name: str, trading_pair: str, interval: str, max_records: int = 500):
        """
        Retrieves the candles for a trading pair from the specified connector.
        :param connector_name: str
        :param trading_pair: str
        :param interval: str
        :param max_records: int
        :return: Candles dataframe.
        """
        candles_df = self.candles_feeds.get(f"{connector_name}_{trading_pair}_{interval}")
        return candles_df[(candles_df["timestamp"] >= self.start_time) & (candles_df["timestamp"] <= self.end_time)]

    def get_price_by_type(self, connector_name: str, trading_pair: str, price_type: PriceType):
        """
        Retrieves the price for a trading pair from the specified connector based on the price type.
        :param connector_name: str
        :param trading_pair: str
        :param price_type: PriceType
        :return: Price.
        """
        return self.prices.get(f"{connector_name}_{trading_pair}", Decimal("1"))

    def quantize_order_amount(self, connector_name: str, trading_pair: str, amount: Decimal):
        """
        Quantizes the order amount based on the trading pair's minimum order size.
        :param connector_name: str
        :param trading_pair: str
        :param amount: Decimal
        :return: Quantized amount.
        """
        trading_rules = self.get_trading_rules(connector_name, trading_pair)
        order_size_quantum = trading_rules.min_base_amount_increment
        return (amount // order_size_quantum) * order_size_quantum

    def quantize_order_price(self, connector_name: str, trading_pair: str, price: Decimal):
        """
        Quantizes the order price based on the trading pair's minimum price increment.
        :param connector_name: str
        :param trading_pair: str
        :param price: Decimal
        :return: Quantized price.
        """
        trading_rules = self.get_trading_rules(connector_name, trading_pair)
        price_quantum = trading_rules.min_price_increment
        return (price // price_quantum) * price_quantum



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: market_data.proto
# Protobuf Python Version: 4.25.2
"""Generated protocol buffer code."""
from google.protobuf import (
    descriptor as _descriptor,
    descriptor_pool as _descriptor_pool,
    symbol_database as _symbol_database,
)
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x11market_data.proto\x12\x0bmarket_data"\x86\x03\n\tMdMessage\x12+\n\theartbeat\x18\x01 \x01(\x0b\x32\x16.market_data.HeartbeatH\x00\x12\'\n\x07summary\x18\x02 \x01(\x0b\x32\x14.market_data.SummaryH\x00\x12%\n\x06trades\x18\x03 \x01(\x0b\x32\x13.market_data.TradesH\x00\x12\x32\n\x0cmbo_snapshot\x18\x04 \x01(\x0b\x32\x1a.market_data.MarketByOrderH\x00\x12\x32\n\x08mbo_diff\x18\x05 \x01(\x0b\x32\x1e.market_data.MarketByOrderDiffH\x00\x12\x32\n\x0cmbp_snapshot\x18\x06 \x01(\x0b\x32\x1a.market_data.MarketByPriceH\x00\x12\x32\n\x08mbp_diff\x18\x07 \x01(\x0b\x32\x1e.market_data.MarketByPriceDiffH\x00\x12#\n\x05kline\x18\x08 \x01(\x0b\x32\x12.market_data.KlineH\x00\x42\x07\n\x05inner"\xaf\x01\n\rMarketByPrice\x12\x30\n\x06levels\x18\x01 \x03(\x0b\x32 .market_data.MarketByPrice.Level\x12\r\n\x05\x63hunk\x18\x02 \x01(\r\x12\x12\n\nnum_chunks\x18\x03 \x01(\r\x1aI\n\x05Level\x12\r\n\x05price\x18\x01 \x01(\x04\x12\x10\n\x08quantity\x18\x02 \x01(\x04\x12\x1f\n\x04side\x18\x03 \x01(\x0e\x32\x11.market_data.Side"\xa4\x02\n\x11MarketByPriceDiff\x12\x32\n\x05\x64iffs\x18\x01 \x03(\x0b\x32#.market_data.MarketByPriceDiff.Diff\x12\x18\n\x10total_bid_levels\x18\x02 \x01(\r\x12\x18\n\x10total_ask_levels\x18\x03 \x01(\r\x1a{\n\x04\x44iff\x12\r\n\x05price\x18\x01 \x01(\x04\x12\x10\n\x08quantity\x18\x02 \x01(\x04\x12\x1f\n\x04side\x18\x03 \x01(\x0e\x32\x11.market_data.Side\x12\x31\n\x02op\x18\x04 \x01(\x0e\x32%.market_data.MarketByPriceDiff.DiffOp"*\n\x06\x44iffOp\x12\x07\n\x03\x41\x44\x44\x10\x00\x12\n\n\x06REMOVE\x10\x01\x12\x0b\n\x07REPLACE\x10\x02"\xdc\x01\n\rMarketByOrder\x12\x30\n\x06orders\x18\x01 \x03(\x0b\x32 .market_data.MarketByOrder.Order\x12\r\n\x05\x63hunk\x18\x02 \x01(\r\x12\x12\n\nnum_chunks\x18\x03 \x01(\r\x1av\n\x05Order\x12\r\n\x05price\x18\x01 \x01(\x04\x12\x10\n\x08quantity\x18\x02 \x01(\x04\x12\x19\n\x11\x65xchange_order_id\x18\x03 \x01(\x04\x12\x1f\n\x04side\x18\x04 \x01(\x0e\x32\x11.market_data.Side\x12\x10\n\x08priority\x18\x05 \x01(\x04"\x86\x03\n\x11MarketByOrderDiff\x12\x32\n\x05\x64iffs\x18\x01 \x03(\x0b\x32#.market_data.MarketByOrderDiff.Diff\x12\x18\n\x10total_bid_levels\x18\x02 \x01(\r\x12\x18\n\x10total_ask_levels\x18\x03 \x01(\r\x12\x18\n\x10total_bid_orders\x18\x04 \x01(\r\x12\x18\n\x10total_ask_orders\x18\x05 \x01(\r\x1a\xa8\x01\n\x04\x44iff\x12\r\n\x05price\x18\x01 \x01(\x04\x12\x10\n\x08quantity\x18\x02 \x01(\x04\x12\x19\n\x11\x65xchange_order_id\x18\x03 \x01(\x04\x12\x1f\n\x04side\x18\x04 \x01(\x0e\x32\x11.market_data.Side\x12\x31\n\x02op\x18\x05 \x01(\x0e\x32%.market_data.MarketByOrderDiff.DiffOp\x12\x10\n\x08priority\x18\x06 \x01(\x04"*\n\x06\x44iffOp\x12\x07\n\x03\x41\x44\x44\x10\x00\x12\n\n\x06REMOVE\x10\x01\x12\x0b\n\x07REPLACE\x10\x02"\x80\x02\n\x06Trades\x12)\n\x06trades\x18\x01 \x03(\x0b\x32\x19.market_data.Trades.Trade\x1a\xca\x01\n\x05Trade\x12\x0f\n\x07tradeId\x18\x01 \x01(\x04\x12\r\n\x05price\x18\x02 \x01(\x04\x12*\n\x0f\x61ggressing_side\x18\x03 \x01(\x0e\x32\x11.market_data.Side\x12!\n\x19resting_exchange_order_id\x18\x04 \x01(\x04\x12\x15\n\rfill_quantity\x18\x05 \x01(\x04\x12\x15\n\rtransact_time\x18\x06 \x01(\x04\x12$\n\x1c\x61ggressing_exchange_order_id\x18\x07 \x01(\x04"\xdb\x01\n\x07Summary\x12\x11\n\x04open\x18\x01 \x01(\x04H\x00\x88\x01\x01\x12\x12\n\x05\x63lose\x18\x02 \x01(\x04H\x01\x88\x01\x01\x12\x10\n\x03low\x18\x03 \x01(\x04H\x02\x88\x01\x01\x12\x11\n\x04high\x18\x04 \x01(\x04H\x03\x88\x01\x01\x12\x16\n\x0e\x62\x61se_volume_lo\x18\x05 \x01(\x04\x12\x16\n\x0e\x62\x61se_volume_hi\x18\x06 \x01(\x04\x12\x17\n\x0fquote_volume_lo\x18\x07 \x01(\x04\x12\x17\n\x0fquote_volume_hi\x18\x08 \x01(\x04\x42\x07\n\x05_openB\x08\n\x06_closeB\x06\n\x04_lowB\x07\n\x05_high"\xdf\x01\n\x05Kline\x12,\n\x08interval\x18\x01 \x01(\x0e\x32\x1a.market_data.KlineInterval\x12\x12\n\nstart_time\x18\x02 \x01(\x04\x12\x11\n\x04open\x18\x03 \x01(\x04H\x00\x88\x01\x01\x12\x12\n\x05\x63lose\x18\x04 \x01(\x04H\x01\x88\x01\x01\x12\x11\n\x04high\x18\x05 \x01(\x04H\x02\x88\x01\x01\x12\x10\n\x03low\x18\x06 \x01(\x04H\x03\x88\x01\x01\x12\x11\n\tvolume_lo\x18\x07 \x01(\x04\x12\x11\n\tvolume_hi\x18\x08 \x01(\x04\x42\x07\n\x05_openB\x08\n\x06_closeB\x07\n\x05_highB\x06\n\x04_low"2\n\tHeartbeat\x12\x12\n\nrequest_id\x18\x01 \x01(\x04\x12\x11\n\ttimestamp\x18\x02 \x01(\x04"6\n\nMdMessages\x12(\n\x08messages\x18\x01 \x03(\x0b\x32\x16.market_data.MdMessage"\xa5\x01\n\nAggMessage\x12+\n\theartbeat\x18\x01 \x01(\x0b\x32\x16.market_data.HeartbeatH\x00\x12/\n\x0ctop_of_books\x18\x02 \x01(\x0b\x32\x17.market_data.TopOfBooksH\x00\x12\x30\n\x0crate_updates\x18\x03 \x01(\x0b\x32\x18.market_data.RateUpdatesH\x00\x42\x07\n\x05inner"\xb5\x02\n\tTopOfBook\x12\x11\n\tmarket_id\x18\x01 \x01(\x04\x12\x15\n\rtransact_time\x18\x02 \x01(\x04\x12\x16\n\tbid_price\x18\x03 \x01(\x04H\x00\x88\x01\x01\x12\x19\n\x0c\x62id_quantity\x18\x04 \x01(\x04H\x01\x88\x01\x01\x12\x16\n\task_price\x18\x05 \x01(\x04H\x02\x88\x01\x01\x12\x19\n\x0c\x61sk_quantity\x18\x06 \x01(\x04H\x03\x88\x01\x01\x12\x17\n\nlast_price\x18\x07 \x01(\x04H\x04\x88\x01\x01\x12\x1d\n\x10rolling24h_price\x18\x08 \x01(\x04H\x05\x88\x01\x01\x42\x0c\n\n_bid_priceB\x0f\n\r_bid_quantityB\x0c\n\n_ask_priceB\x0f\n\r_ask_quantityB\r\n\x0b_last_priceB\x13\n\x11_rolling24h_price"2\n\nTopOfBooks\x12$\n\x04tops\x18\x01 \x03(\x0b\x32\x16.market_data.TopOfBook"j\n\nRateUpdate\x12\x10\n\x08\x61sset_id\x18\x01 \x01(\x04\x12\x11\n\ttimestamp\x18\x02 \x01(\x04\x12\x0c\n\x04rate\x18\x03 \x01(\x04\x12)\n\x04side\x18\x04 \x01(\x0e\x32\x1b.market_data.RateUpdateSide"7\n\x0bRateUpdates\x12(\n\x07updates\x18\x01 \x03(\x0b\x32\x17.market_data.RateUpdate"l\n\rClientMessage\x12+\n\theartbeat\x18\x01 \x01(\x0b\x32\x16.market_data.HeartbeatH\x00\x12%\n\x06\x63onfig\x18\x02 \x01(\x0b\x32\x13.market_data.ConfigH\x00\x42\x07\n\x05inner"o\n\x06\x43onfig\x12\x0b\n\x03mbp\x18\x01 \x01(\x08\x12\x0b\n\x03mbo\x18\x02 \x01(\x08\x12\x0e\n\x06trades\x18\x03 \x01(\x08\x12\x0f\n\x07summary\x18\x04 \x01(\x08\x12*\n\x06klines\x18\x05 \x03(\x0e\x32\x1a.market_data.KlineInterval*\x18\n\x04Side\x12\x07\n\x03\x42ID\x10\x00\x12\x07\n\x03\x41SK\x10\x01*@\n\rKlineInterval\x12\x06\n\x02S1\x10\x00\x12\x06\n\x02M1\x10\x01\x12\x07\n\x03M15\x10\x02\x12\x06\n\x02H1\x10\x03\x12\x06\n\x02H4\x10\x04\x12\x06\n\x02\x44\x31\x10\x05*%\n\x0eRateUpdateSide\x12\x08\n\x04\x42\x41SE\x10\x00\x12\t\n\x05QUOTE\x10\x01\x42\x17Z\x03go/\xaa\x02\x0f\x43ube.MarketDatab\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "market_data_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS is False:
    _globals["DESCRIPTOR"]._options = None
    _globals["DESCRIPTOR"]._serialized_options = b"Z\003go/\252\002\017Cube.MarketData"
    _globals["_SIDE"]._serialized_start = 3251
    _globals["_SIDE"]._serialized_end = 3275
    _globals["_KLINEINTERVAL"]._serialized_start = 3277
    _globals["_KLINEINTERVAL"]._serialized_end = 3341
    _globals["_RATEUPDATESIDE"]._serialized_start = 3343
    _globals["_RATEUPDATESIDE"]._serialized_end = 3380
    _globals["_MDMESSAGE"]._serialized_start = 35
    _globals["_MDMESSAGE"]._serialized_end = 425
    _globals["_MARKETBYPRICE"]._serialized_start = 428
    _globals["_MARKETBYPRICE"]._serialized_end = 603
    _globals["_MARKETBYPRICE_LEVEL"]._serialized_start = 530
    _globals["_MARKETBYPRICE_LEVEL"]._serialized_end = 603
    _globals["_MARKETBYPRICEDIFF"]._serialized_start = 606
    _globals["_MARKETBYPRICEDIFF"]._serialized_end = 898
    _globals["_MARKETBYPRICEDIFF_DIFF"]._serialized_start = 731
    _globals["_MARKETBYPRICEDIFF_DIFF"]._serialized_end = 854
    _globals["_MARKETBYPRICEDIFF_DIFFOP"]._serialized_start = 856
    _globals["_MARKETBYPRICEDIFF_DIFFOP"]._serialized_end = 898
    _globals["_MARKETBYORDER"]._serialized_start = 901
    _globals["_MARKETBYORDER"]._serialized_end = 1121
    _globals["_MARKETBYORDER_ORDER"]._serialized_start = 1003
    _globals["_MARKETBYORDER_ORDER"]._serialized_end = 1121
    _globals["_MARKETBYORDERDIFF"]._serialized_start = 1124
    _globals["_MARKETBYORDERDIFF"]._serialized_end = 1514
    _globals["_MARKETBYORDERDIFF_DIFF"]._serialized_start = 1302
    _globals["_MARKETBYORDERDIFF_DIFF"]._serialized_end = 1470
    _globals["_MARKETBYORDERDIFF_DIFFOP"]._serialized_start = 856
    _globals["_MARKETBYORDERDIFF_DIFFOP"]._serialized_end = 898
    _globals["_TRADES"]._serialized_start = 1517
    _globals["_TRADES"]._serialized_end = 1773
    _globals["_TRADES_TRADE"]._serialized_start = 1571
    _globals["_TRADES_TRADE"]._serialized_end = 1773
    _globals["_SUMMARY"]._serialized_start = 1776
    _globals["_SUMMARY"]._serialized_end = 1995
    _globals["_KLINE"]._serialized_start = 1998
    _globals["_KLINE"]._serialized_end = 2221
    _globals["_HEARTBEAT"]._serialized_start = 2223
    _globals["_HEARTBEAT"]._serialized_end = 2273
    _globals["_MDMESSAGES"]._serialized_start = 2275
    _globals["_MDMESSAGES"]._serialized_end = 2329
    _globals["_AGGMESSAGE"]._serialized_start = 2332
    _globals["_AGGMESSAGE"]._serialized_end = 2497
    _globals["_TOPOFBOOK"]._serialized_start = 2500
    _globals["_TOPOFBOOK"]._serialized_end = 2809
    _globals["_TOPOFBOOKS"]._serialized_start = 2811
    _globals["_TOPOFBOOKS"]._serialized_end = 2861
    _globals["_RATEUPDATE"]._serialized_start = 2863
    _globals["_RATEUPDATE"]._serialized_end = 2969
    _globals["_RATEUPDATES"]._serialized_start = 2971
    _globals["_RATEUPDATES"]._serialized_end = 3026
    _globals["_CLIENTMESSAGE"]._serialized_start = 3028
    _globals["_CLIENTMESSAGE"]._serialized_end = 3136
    _globals["_CONFIG"]._serialized_start = 3138
    _globals["_CONFIG"]._serialized_end = 3249
# @@protoc_insertion_point(module_scope)



import asyncio
import time
from collections import defaultdict
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Dict, List, Mapping, Optional

import hummingbot.connector.derivative.binance_perpetual.binance_perpetual_constants as CONSTANTS
import hummingbot.connector.derivative.binance_perpetual.binance_perpetual_web_utils as web_utils
from hummingbot.core.data_type.common import TradeType
from hummingbot.core.data_type.funding_info import FundingInfo, FundingInfoUpdate
from hummingbot.core.data_type.order_book_message import OrderBookMessage, OrderBookMessageType
from hummingbot.core.data_type.perpetual_api_order_book_data_source import PerpetualAPIOrderBookDataSource
from hummingbot.core.web_assistant.connections.data_types import WSJSONRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.derivative.binance_perpetual.binance_perpetual_derivative import (
        BinancePerpetualDerivative,
    )


class BinancePerpetualAPIOrderBookDataSource(PerpetualAPIOrderBookDataSource):
    _bpobds_logger: Optional[HummingbotLogger] = None
    _trading_pair_symbol_map: Dict[str, Mapping[str, str]] = {}
    _mapping_initialization_lock = asyncio.Lock()

    def __init__(
            self,
            trading_pairs: List[str],
            connector: 'BinancePerpetualDerivative',
            api_factory: WebAssistantsFactory,
            domain: str = CONSTANTS.DOMAIN
    ):
        super().__init__(trading_pairs)
        self._connector = connector
        self._api_factory = api_factory
        self._domain = domain
        self._trading_pairs: List[str] = trading_pairs
        self._message_queue: Dict[str, asyncio.Queue] = defaultdict(asyncio.Queue)
        self._trade_messages_queue_key = CONSTANTS.TRADE_STREAM_ID
        self._diff_messages_queue_key = CONSTANTS.DIFF_STREAM_ID
        self._funding_info_messages_queue_key = CONSTANTS.FUNDING_INFO_STREAM_ID
        self._snapshot_messages_queue_key = "order_book_snapshot"

    async def get_last_traded_prices(self,
                                     trading_pairs: List[str],
                                     domain: Optional[str] = None) -> Dict[str, float]:
        return await self._connector.get_last_traded_prices(trading_pairs=trading_pairs)

    async def get_funding_info(self, trading_pair: str) -> FundingInfo:
        symbol_info: Dict[str, Any] = await self._request_complete_funding_info(trading_pair)
        funding_info = FundingInfo(
            trading_pair=trading_pair,
            index_price=Decimal(symbol_info["indexPrice"]),
            mark_price=Decimal(symbol_info["markPrice"]),
            next_funding_utc_timestamp=int(float(symbol_info["nextFundingTime"]) * 1e-3),
            rate=Decimal(symbol_info["lastFundingRate"]),
        )
        return funding_info

    async def _request_order_book_snapshot(self, trading_pair: str) -> Dict[str, Any]:
        ex_trading_pair = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)

        params = {
            "symbol": ex_trading_pair,
            "limit": "1000"
        }

        data = await self._connector._api_get(
            path_url=CONSTANTS.SNAPSHOT_REST_URL,
            params=params)
        return data

    async def _order_book_snapshot(self, trading_pair: str) -> OrderBookMessage:
        snapshot_response: Dict[str, Any] = await self._request_order_book_snapshot(trading_pair)
        snapshot_timestamp: float = time.time()
        snapshot_response.update({"trading_pair": trading_pair})
        snapshot_msg: OrderBookMessage = OrderBookMessage(OrderBookMessageType.SNAPSHOT, {
            "trading_pair": snapshot_response["trading_pair"],
            "update_id": snapshot_response["lastUpdateId"],
            "bids": snapshot_response["bids"],
            "asks": snapshot_response["asks"]
        }, timestamp=snapshot_timestamp)
        return snapshot_msg

    async def _connected_websocket_assistant(self) -> WSAssistant:
        url = f"{web_utils.wss_url(CONSTANTS.PUBLIC_WS_ENDPOINT, self._domain)}"
        ws: WSAssistant = await self._api_factory.get_ws_assistant()
        await ws.connect(ws_url=url, ping_timeout=CONSTANTS.HEARTBEAT_TIME_INTERVAL)
        return ws

    async def _subscribe_channels(self, ws: WSAssistant):
        """
        Subscribes to the trade events and diff orders events through the provided websocket connection.
        :param ws: the websocket assistant used to connect to the exchange
        """
        try:
            stream_id_channel_pairs = [
                (CONSTANTS.DIFF_STREAM_ID, "@depth"),
                (CONSTANTS.TRADE_STREAM_ID, "@aggTrade"),
                (CONSTANTS.FUNDING_INFO_STREAM_ID, "@markPrice"),
            ]
            for stream_id, channel in stream_id_channel_pairs:
                params = []
                for trading_pair in self._trading_pairs:
                    symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
                    params.append(f"{symbol.lower()}{channel}")
                payload = {
                    "method": "SUBSCRIBE",
                    "params": params,
                    "id": stream_id,
                }
                subscribe_request: WSJSONRequest = WSJSONRequest(payload)
                await ws.send(subscribe_request)
            self.logger().info("Subscribed to public order book, trade and funding info channels...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception("Unexpected error occurred subscribing to order book trading and delta streams...")
            raise

    def _channel_originating_message(self, event_message: Dict[str, Any]) -> str:
        channel = ""
        if "result" not in event_message:
            stream_name = event_message.get("stream")
            if "@depth" in stream_name:
                channel = self._diff_messages_queue_key
            elif "@aggTrade" in stream_name:
                channel = self._trade_messages_queue_key
            elif "@markPrice" in stream_name:
                channel = self._funding_info_messages_queue_key
        return channel

    async def _parse_order_book_diff_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        timestamp: float = time.time()
        raw_message["data"]["s"] = await self._connector.trading_pair_associated_to_exchange_symbol(
            raw_message["data"]["s"])
        data = raw_message["data"]
        order_book_message: OrderBookMessage = OrderBookMessage(OrderBookMessageType.DIFF, {
            "trading_pair": data["s"],
            "update_id": data["u"],
            "bids": data["b"],
            "asks": data["a"]
        }, timestamp=timestamp)
        message_queue.put_nowait(order_book_message)

    async def _parse_trade_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        raw_message["data"]["s"] = await self._connector.trading_pair_associated_to_exchange_symbol(
            raw_message["data"]["s"])
        data = raw_message["data"]
        trade_message: OrderBookMessage = OrderBookMessage(OrderBookMessageType.TRADE, {
            "trading_pair": data["s"],
            "trade_type": float(TradeType.SELL.value) if data["m"] else float(TradeType.BUY.value),
            "trade_id": data["a"],
            "update_id": data["E"],
            "price": data["p"],
            "amount": data["q"]
        }, timestamp=data["E"] * 1e-3)

        message_queue.put_nowait(trade_message)

    async def listen_for_order_book_snapshots(self, ev_loop: asyncio.BaseEventLoop, output: asyncio.Queue):
        while True:
            try:
                for trading_pair in self._trading_pairs:
                    snapshot_msg: OrderBookMessage = await self._order_book_snapshot(trading_pair)
                    output.put_nowait(snapshot_msg)
                    self.logger().debug(f"Saved order book snapshot for {trading_pair}")
                delta = CONSTANTS.ONE_HOUR - time.time() % CONSTANTS.ONE_HOUR
                await self._sleep(delta)
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().error(
                    "Unexpected error occurred fetching orderbook snapshots. Retrying in 5 seconds...", exc_info=True
                )
                await self._sleep(5.0)

    async def _parse_funding_info_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):

        data: Dict[str, Any] = raw_message["data"]
        trading_pair = await self._connector.trading_pair_associated_to_exchange_symbol(data["s"])

        if trading_pair not in self._trading_pairs:
            return
        funding_info = FundingInfoUpdate(
            trading_pair=trading_pair,
            index_price=Decimal(data["i"]),
            mark_price=Decimal(data["p"]),
            next_funding_utc_timestamp=int(float(data["T"]) * 1e-3),
            rate=Decimal(data["r"]),
        )

        message_queue.put_nowait(funding_info)

    async def _request_complete_funding_info(self, trading_pair: str):
        ex_trading_pair = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        data = await self._connector._api_get(
            path_url=CONSTANTS.MARK_PRICE_URL,
            params={"symbol": ex_trading_pair},
            is_auth_required=True)
        return data
